import abc
import collections
import json
import pydoc

import six
import terminaltables
from click import style
from halo import halo

from gradient import api_sdk, exceptions
from gradient.api_sdk import sdk_exceptions
from gradient.api_sdk.config import config
from gradient.clilogger import CliLogger
from gradient.cliutils import get_terminal_lines, TerminalPrinter
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.exceptions import ApplicationError


@six.add_metaclass(abc.ABCMeta)
class BaseCommand:
    def __init__(self, api_key, logger=CliLogger()):
        self.api_key = api_key
        self.client = self._get_client(api_key, logger)
        self.logger = logger

    def get_namespace(self):
        client = api_sdk.clients.http_client.API(
            config.CONFIG_HOST,
            api_key=self.api_key,
            logger=self.logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )

        try:
            resp = client.get(url='/teams/namespace')
        except sdk_exceptions.GradientSdkError as e:
                raise exceptions.ReceivingDataFailedError(e)
        if isinstance(resp.json(), str):
            return resp.json()
        return ''

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def _get_client(self, api_key, logger):
        pass


@six.add_metaclass(abc.ABCMeta)
class ListCommandMixin(object):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."
    TOTAL_ITEMS_KEY = "total"

    def execute(self, **kwargs):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instances = self._get_instances(kwargs)

        self._log_objects_list(instances)

    @abc.abstractmethod
    def _get_instances(self, kwargs):
        pass

    @abc.abstractmethod
    def _get_table_data(self, objects):
        pass

    def _log_objects_list(self, objects):
        if not objects:
            self.logger.warning("No data found")
            return

        table_data = self._get_table_data(objects)
        table_str = self._make_list_table(table_data)
        self._print_table_to_terminal(table_str)

    def _print_table_to_terminal(self, table_str):
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    @staticmethod
    def _make_list_table(table_data):
        ascii_table = terminaltables.AsciiTable(table_data)
        table_string = ascii_table.table
        return table_string

    def _generate_data_table(self, **kwargs):
        limit = kwargs.get("limit")
        offset = kwargs.get("offset")
        meta_data = dict()
        while self.TOTAL_ITEMS_KEY not in meta_data or offset < meta_data.get(self.TOTAL_ITEMS_KEY):
            with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
                kwargs["offset"] = offset
                instances, meta_data = self._get_instances(
                    **kwargs
                )
            next_iteration = False
            if instances:
                table_data = self._get_table_data(instances)
                table_str = self._make_list_table(table_data) + "\n"
                if offset + limit < meta_data.get(self.TOTAL_ITEMS_KEY):
                    next_iteration = True
            else:
                table_str = "No data found"

            yield table_str, next_iteration
            offset += limit


@six.add_metaclass(abc.ABCMeta)
class ListCommandPagerMixin(ListCommandMixin):
    def _get_instances(self, kwargs):
        limit = kwargs.get('limit') or 20
        offset = 0

        while True:
            with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
                results = self.client.list(limit=limit + 1, offset=offset, **kwargs)
            has_more = len(results) > limit
            yield results[:limit], has_more
            if not has_more:
                break
            offset += limit

    def execute(self, **kwargs):
        for instances, has_more in self._get_instances(kwargs):
            self._log_objects_list(instances)
            yield has_more


@six.add_metaclass(abc.ABCMeta)
class DetailsCommandMixin(object):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, id_):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instance = self._get_instance(id_)

        self._log_object(instance)

    def _get_instance(self, id_):
        instance = self.client.get(id_)

        return instance

    def _log_object(self, instance):
        if instance is None:
            self.logger.warning("Not found")
            return

        table_str = self._make_table(instance)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    def _make_table(self, instance):
        data = self._get_table_data(instance)
        ascii_table = terminaltables.AsciiTable(data)
        table_string = ascii_table.table
        return table_string

    @abc.abstractmethod
    def _get_table_data(self, instance):
        pass


class StreamMetricsCommand(ListCommandMixin):
    def __init__(self, *args, **kwargs):
        super(StreamMetricsCommand, self).__init__(*args, **kwargs)
        # {"metricName": {"pod_id": "value"}}
        self._recent_values = collections.OrderedDict()
        self.terminal_printer = None

    def execute(self, *args, **kwargs):
        self.terminal_printer = TerminalPrinter()
        self.terminal_printer.init()
        try:
            rv = super(StreamMetricsCommand, self).execute(*args, **kwargs)
        finally:
            self.terminal_printer.clear()
            self.terminal_printer.cleanup()

        return rv

    def _get_instances(self, kwargs):
        metrics_stream = self.client.stream_metrics(**kwargs)
        self._prepare_recent_values(kwargs["built_in_metrics"])
        return metrics_stream

    def _prepare_recent_values(self, builtin_metrics_names):
        for name in builtin_metrics_names:
            self._recent_values[name] = collections.OrderedDict()

    def _log_objects_list(self, metrics_stream):
        for raw_metric_date_response in metrics_stream:
            metric_data = json.loads(raw_metric_date_response)
            self._update_recent_values(metric_data)
            super(StreamMetricsCommand, self)._log_objects_list(raw_metric_date_response)

    def _update_recent_values(self, metric_data):
        metric_name = metric_data["chart_name"]
        metrics = metric_data["pod_metrics"]
        for pod_name, data in metrics.items():
            self._recent_values[metric_name][pod_name] = data["value"]

    def _print_table_to_terminal(self, table_str):
        self.terminal_printer.rewrite_screen(table_str)

    def _get_table_data(self, objects):
        metrics = list(self._recent_values.keys())
        table_ = ["Pod"] + metrics
        table_data = [table_]
        values = collections.OrderedDict()  # {pod_name: {metricName: value}}
        for metric_name, data in self._recent_values.items():
            for pod_name, value in data.items():
                pod_metrics = values.setdefault(pod_name, collections.OrderedDict())
                pod_metrics[metric_name] = value

        for pod_name, pod_metrics in sorted(values.items()):
            row = [pod_name]
            for metric_name in metrics:
                value = pod_metrics.get(metric_name, "")
                row.append(value)

            table_data.append(row)

        return table_data


class LogsCommandMixin(object):
    def execute(self, id, line, limit, follow):
        if follow:
            self.logger.log("Awaiting logs...")
            self._log_logs_continuously(id, line, limit)
        else:
            self._log_table_of_logs(id, line, limit)

    def _log_table_of_logs(self, id, line, limit):
        logs = self.client.logs(id, line, limit)
        if not logs:
            raise ApplicationError("No logs found")

        table_str = self._make_table(logs, id)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    def _log_logs_continuously(self, id, line, limit):
        logs_gen = self._get_logs_generator(id, line, limit)
        for log in logs_gen:
            log_msg = self._get_log_row_string(id, log)
            self.logger.log(log_msg)

    def _get_logs_generator(self, id, line, limit):
        logs_gen = self.client.yield_logs(id, line, limit)
        return logs_gen

    def _make_table(self, logs, id):
        table_title = "%s %s logs" % (self.ENTITY, id)
        table_data = [("LINE", "MESSAGE")]
        table = terminaltables.AsciiTable(table_data, title=table_title)

        for log in logs:
            table_data.append(self._format_row(log))

        return table.table

    def _get_log_row_string(self, id, log):
        log_msg = "{}\t{}".format(*self._format_row(log))
        return log_msg

    @staticmethod
    def _format_row(log_row):
        return (style(fg="red", text=str(log_row.line)),
                log_row.message)



import mock
import pytest

from gradient import version_checker
from gradient.version import version


@mock.patch("gradient.version_checker.GradientVersionChecker._should_check_version")
@mock.patch("gradient.version_checker.VersionChecker.is_up_to_date")
@mock.patch("gradient.version_checker.logger")
def test_should_check_for_new_gradient_version_and_print_proper_warning_when_current_version_is_old(
        logger_patched, is_up_to_date_patched, should_check_patched):
    should_check_patched.return_value = True
    is_up_to_date_patched.return_value = (False, "1.2.3")

    version_checker.GradientVersionChecker.look_for_new_version()

    is_up_to_date_patched.assert_called_once()
    logger_patched.warning.assert_called_once_with(
        "Warning: this version of the Gradient CLI ({}) is out of date. "
        "Some functionality might not be supported until you upgrade. \n\n"
        "Run `pip install -U gradient` to upgrade\n".format(version))


@mock.patch("gradient.version_checker.GradientVersionChecker._should_check_version")
@mock.patch("gradient.version_checker.VersionChecker.is_up_to_date")
@mock.patch("gradient.version_checker.logger")
def test_should_check_for_new_gradient_version_and_not_print_anything_when_current_version_is_latest(
        logger_patched, is_up_to_date_patched, should_check_patched):
    should_check_patched.return_value = True
    is_up_to_date_patched.return_value = (True, "1.2.3")

    version_checker.GradientVersionChecker.look_for_new_version()

    is_up_to_date_patched.assert_called_once()
    logger_patched.warning.assert_not_called()


@mock.patch("gradient.version_checker.xmlrpclib.ServerProxy")
def test_should_return_package_version_when_get_version_was_run(sever_proxy_class_patched):
    pypi_patched = mock.MagicMock()
    pypi_patched.package_releases.return_value = ["1.2.3"]
    sever_proxy_class_patched.return_value = pypi_patched

    vc = version_checker.VersionChecker()
    latest_version = vc.get_version_from_repository("some_module_name")

    sever_proxy_class_patched.assert_called_with("http://pypi.python.org/pypi")
    assert latest_version == "1.2.3"


@mock.patch("gradient.version_checker.xmlrpclib.ServerProxy")
def test_should_raise_proper_exception_when_get_version_was_run_and_package_was_not_found(sever_proxy_class_patched):
    pypi_patched = mock.MagicMock()
    pypi_patched.package_releases.return_value = []
    sever_proxy_class_patched.return_value = pypi_patched

    vc = version_checker.VersionChecker()
    with pytest.raises(version_checker.PackageNotFoundError):
        vc.get_version_from_repository("some_module_name")

    sever_proxy_class_patched.assert_called_with("http://pypi.python.org/pypi")


@mock.patch("gradient.version_checker.VersionChecker.get_version_from_repository")
def test_should_return_true_when_current_version_equals_latest_from_pypi(get_version_patched):
    get_version_patched.return_value = "1.2.3"

    vc = version_checker.VersionChecker()
    up_to_date, version_in_repository = vc.is_up_to_date("some_module_name", "1.2.3")

    assert up_to_date
    assert version_in_repository == "1.2.3"


@mock.patch("gradient.version_checker.VersionChecker.get_version_from_repository")
def test_should_return_true_when_current_version_is_higher_than_latest_from_pypi(get_version_patched):
    get_version_patched.return_value = "1.2.3"

    vc = version_checker.VersionChecker()
    up_to_date, version_in_repository = vc.is_up_to_date("some_module_name", "1.2.4a0")

    assert up_to_date
    assert version_in_repository == "1.2.3"


@mock.patch("gradient.version_checker.VersionChecker.get_version_from_repository")
def test_should_return_false_when_current_version_is_lower_than_latest_from_pypi(get_version_patched):
    get_version_patched.return_value = "1.2.3"

    vc = version_checker.VersionChecker()
    up_to_date, version_in_repository = vc.is_up_to_date("some_module_name", "1.2.1")

    assert not up_to_date
    assert version_in_repository == "1.2.3"



import mock
from click.testing import CliRunner

from gradient.api_sdk.clients import http_client
from gradient.cli import cli
from tests import example_responses, MockResponse

EXPECTED_HEADERS = http_client.default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestClustersList(object):
    URL = "https://api.paperspace.io/clusters/getClusters"

    COMMAND = ["clusters", "list"]
    COMMAND_WITH_LIMIT = COMMAND[:] + ["-l", "2"]
    COMMAND_WITH_OPTIONS_FILE = ["clusters", "list", "--optionsFile", ]  # path added in test

    LIST_CLUSTERS = example_responses.EXAMPLE_CLUSTERS_LIST_RESPONSE
    LIMITED_LIST_CLUSTERS = example_responses.LIMITED_EXAMPLE_CLUSTERS_LIST_RESPONSE

    DEFAULT_PARAMS = {
        "filter": '{"limit": 20, "offset": 0, "where": {"isPrivate": true}}'
    }
    LIMITED_PARAMS = {
        "filter": '{"limit": 2, "offset": 0, "where": {"isPrivate": true}}'
    }

    COMMAND_WITH_API_KEY = ["clusters", "list", "--apiKey", "some_key"]

    LIST_STDOUT = """+--------------+----------------+----------------------------+
| ID           | Name           | Type                       |
+--------------+----------------+----------------------------+
| cluster_id_1 | cluster name 1 | Job Cluster                |
| cluster_id_2 | cluster name 2 | Kubernetes Processing Site |
| cluster_id_3 | cluster name 3 | Job Cluster                |
+--------------+----------------+----------------------------+
"""
    PAGINATED_LIST_STDOUT = """+--------------+----------------+----------------------------+
| ID           | Name           | Type                       |
+--------------+----------------+----------------------------+
| cluster_id_1 | cluster name 1 | Job Cluster                |
| cluster_id_2 | cluster name 2 | Kubernetes Processing Site |
+--------------+----------------+----------------------------+

Do you want to continue? [y/N]: 
Aborted!
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_clusters(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_CLUSTERS)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert self.LIST_STDOUT in result.output, result.exc_info
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=self.DEFAULT_PARAMS)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_paginate_list_of_clusters(self, get_patched):
        get_patched.return_value = MockResponse(self.LIMITED_LIST_CLUSTERS)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_LIMIT)

        assert self.PAGINATED_LIST_STDOUT in str(result.output)
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=self.LIMITED_PARAMS)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_defined_in_a_config_file(self, get_patched, clusters_list_config_path):
        get_patched.return_value = MockResponse(json_data=self.LIMITED_LIST_CLUSTERS)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [clusters_list_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert self.PAGINATED_LIST_STDOUT in str(result.output)
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=self.LIMITED_PARAMS,
                                            )


class TestListVmTypes(object):
    URL = "https://api.paperspace.io/vmTypes/getVmTypesByClusters"

    COMMAND = ["clusters", "machineTypes", "list"]
    COMMAND_WITH_FILTERING_BY_CLUSTER_ID = ["clusters", "machineTypes", "list", "--clusterId", "cluwffvkb"]
    COMMAND_WITH_FILTERING_BY_INVALID_ID = ["clusters", "machineTypes", "list", "--clusterId", "invalid_id"]
    COMMAND_WITH_OPTIONS_FILE = ["clusters", "machineTypes", "list", "--optionsFile", ]  # path added in test

    LIST_JSON = example_responses.LIST_OF_VM_MACHINE_TYPES
    EXPECTED_STDOUT = """+--------------+---------+-----------+--------------+-----------+------------+---------------------------------+
| Name         | Kind    | CPU Count | RAM [Bytes]  | GPU Count | GPU Model  | Clusters                        |
+--------------+---------+-----------+--------------+-----------+------------+---------------------------------+
| Wolfpass-CPU | cpu     | 24        | 34359738368  | 1         | N/A        | clfe0kr2p                       |
| c5.24xlarge  | aws-cpu | 94        | 206158430208 | 1         | N/A        | cluwffvkb                       |
| c5.4xlarge   | aws-cpu | 16        | 34359738368  | 1         | N/A        | clqr4b0ox, clrvkwq6l, cluwffvkb |
| c5.xlarge    | aws-cpu | 4         | 8589934592   | 1         | N/A        | clqr4b0ox, clrvkwq6l, cluwffvkb |
| p2.xlarge    | aws-gpu | 4         | 65498251264  | 1         | Tesla K80  | clqr4b0ox, clrvkwq6l            |
| p3.16xlarge  | aws-gpu | 64        | 523986010112 | 8         | Tesla V100 | clqr4b0ox, clrvkwq6l, cluwffvkb |
| p3.2xlarge   | aws-gpu | 8         | 65498251264  | 1         | Tesla V100 | clqr4b0ox, clrvkwq6l, cluwffvkb |
+--------------+---------+-----------+--------------+-----------+------------+---------------------------------+
"""

    DETAILS_STDOUT_WITH_FILTERING_BY_CLUSTER_ID = """+-------------+---------+-----------+--------------+-----------+------------+---------------------------------+
| Name        | Kind    | CPU Count | RAM [Bytes]  | GPU Count | GPU Model  | Clusters                        |
+-------------+---------+-----------+--------------+-----------+------------+---------------------------------+
| c5.24xlarge | aws-cpu | 94        | 206158430208 | 1         | N/A        | cluwffvkb                       |
| c5.4xlarge  | aws-cpu | 16        | 34359738368  | 1         | N/A        | clqr4b0ox, clrvkwq6l, cluwffvkb |
| c5.xlarge   | aws-cpu | 4         | 8589934592   | 1         | N/A        | clqr4b0ox, clrvkwq6l, cluwffvkb |
| p3.16xlarge | aws-gpu | 64        | 523986010112 | 8         | Tesla V100 | clqr4b0ox, clrvkwq6l, cluwffvkb |
| p3.2xlarge  | aws-gpu | 8         | 65498251264  | 1         | Tesla V100 | clqr4b0ox, clrvkwq6l, cluwffvkb |
+-------------+---------+-----------+--------------+-----------+------------+---------------------------------+
"""

    RESPONSE_JSON_WHEN_WRONG_API_KEY_WAS_USED = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED = "Failed to fetch data: Invalid API token\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_machine_types(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_vm_machine_types_filtered_by_cluster_id(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_FILTERING_BY_CLUSTER_ID)

        assert result.output == self.DETAILS_STDOUT_WITH_FILTERING_BY_CLUSTER_ID, result.exc_info
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_machine_types_but_none_found(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_FILTERING_BY_INVALID_ID)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert result.output == "No data found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_WRONG_API_KEY_WAS_USED,
                                                status_code=403)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_defined_in_a_config_file(self, get_patched, vm_machine_types_list_config_path):
        get_patched.return_value = MockResponse(self.LIST_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [vm_machine_types_list_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.DETAILS_STDOUT_WITH_FILTERING_BY_CLUSTER_ID
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params=None)



from . import common
from .. import config
from .. import serializers, models


class GetTensorboardApiUrlMixin(object):
    def _get_api_url(self, **kwargs):
        return config.config.CONFIG_SERVICE_HOST


class ParseTensorboardMixin(object):
    def _parse_object(self, instance_dict, **kwargs):
        machine_instance_dict = instance_dict.get("instance")
        if machine_instance_dict:
            machine_instance = serializers.InstanceSchema().get_instance(machine_instance_dict)
        else:
            machine_instance = models.Instance()

        instance = super(ParseTensorboardMixin, self)._parse_object(instance_dict)
        instance.instance = machine_instance
        return instance


class CreateTensorboard(GetTensorboardApiUrlMixin, common.CreateResource):
    SERIALIZER_CLS = serializers.TensorboardSchema

    def get_request_url(self, **kwargs):
        return "/tensorboards/v1/"

    def _get_id_from_response(self, response):
        handle = response.data["data"]["id"]
        return handle


class GetTensorboard(ParseTensorboardMixin, GetTensorboardApiUrlMixin, common.GetResource):
    SERIALIZER_CLS = serializers.TensorboardDetailSchema

    def get_request_url(self, **kwargs):
        id_ = kwargs["id"]
        return "/tensorboards/v1/{}/".format(id_)

    def _parse_object(self, instance_dict, **kwargs):
        instance_dict = instance_dict["data"]
        instance = super(GetTensorboard, self)._parse_object(instance_dict)
        return instance


class ListTensorboards(ParseTensorboardMixin, GetTensorboardApiUrlMixin, common.ListResources):
    SERIALIZER_CLS = serializers.TensorboardDetailSchema

    def _get_instance_dicts(self, data, **kwargs):
        instance_dicts = data["data"]
        return instance_dicts

    def get_request_url(self, **kwargs):
        return "/tensorboards/v1/"


class UpdateTensorboard(ParseTensorboardMixin, GetTensorboardApiUrlMixin, common.AlterResource):
    VALIDATION_ERROR_MESSAGE = "Failed to update resource"

    def update(self, id, **kwargs):
        resp = self._run(id=id, **kwargs)
        tensorboard = self._parse_object(resp.data)
        return tensorboard

    def get_request_url(self, **kwargs):
        return "/tensorboards/v1/{}".format(kwargs["id"])

    def _send_request(self, client, url, json_data=None):
        response = client.put(url, json=json_data)
        return response

    def _get_request_json(self, kwargs):
        return {
            "added_experiments": kwargs.get("added_experiments", list()),
            "removed_experiments": kwargs.get("removed_experiments", list())
        }

    def _parse_object(self, instance_dict, **kwargs):
        instance_dict = instance_dict["data"]
        instance = self._get_instance(instance_dict)
        return instance

    def _get_instance(self, instance_dict):
        machine_instance_dict = instance_dict.get("instance")
        if machine_instance_dict:
            machine_instance = serializers.InstanceSchema().get_instance(machine_instance_dict)
        else:
            machine_instance = models.Instance()

        instance = serializers.TensorboardDetailSchema().get_instance(instance_dict)
        instance.instance = machine_instance
        return instance


class DeleteTensorboard(GetTensorboardApiUrlMixin, common.DeleteResource):
    VALIDATION_ERROR_MESSAGE = "Failed to delete resource"

    def get_request_url(self, **kwargs):
        return "/tensorboards/v1/{}".format(kwargs["id"])



import abc
import os
import time

import requests
import six

from . import sdk_exceptions
from .clients import JobsClient, ModelsClient
from .clients.base_client import BaseClient
from .logger import MuteLogger


class S3FilesDownloader(object):
    def __init__(self, logger=MuteLogger()):
        self.logger = logger
        self.file_download_retries = 8

    def download_list(self, sources, destination_dir):
        """

        :param tuple[tuple[str,str]] sources: tuple/list of (file_path, file_url) pairs
        :param str destination_dir:
        """
        for source in sources:
            self.download_file(source, destination_dir, max_retries=self.file_download_retries)

    def download_file(self, source, destination_dir, max_retries=0):
        self._create_directory(destination_dir)

        file_path, file_url = source
        self.logger.log("Downloading: {}".format(file_path))

        # Trying to download several times in case of connection error with S3.
        # The error seems to occur randomly but adding short sleep between retries helps a bit
        for _ in range(max_retries + 1):
            try:
                response = requests.get(file_url)
                break
            except requests.exceptions.ConnectionError:
                self.logger.debug("Downloading {} resulted in error. Trying again...".format(file_path))
                time.sleep(0.1)
        else:  # break statement not executed - ConnectionError `max_retries` times
            raise sdk_exceptions.ResourceFetchingError("Downloading {} resulted in error".format(file_path))

        self._create_subdirectories(file_path, destination_dir)
        self._save_file(response, file_path, destination_dir)

    def _create_directory(self, destination_dir):
        if os.path.exists(destination_dir) and os.path.isdir(destination_dir):
            return

        os.makedirs(destination_dir)

    def _create_subdirectories(self, file_path, destination_dir):
        file_dirname = os.path.dirname(file_path)
        file_dir_path = os.path.join(destination_dir, file_dirname)
        self._create_directory(file_dir_path)

    def _save_file(self, response, file_path, destination_dir):
        destination_path = os.path.join(destination_dir, file_path)
        try:
            with open(destination_path, "wb") as h:
                h.write(response.content)
        except TypeError:  # in py3 TypeError is raised when trying to write str in bytes mode so trying in txt mode
            with open(destination_path, "w") as h:
                h.write(response.content)


@six.add_metaclass(abc.ABCMeta)
class ResourceDownloader(object):
    CLIENT_CLASS = None

    def __init__(self, api_key, logger=MuteLogger(), ps_client_name=None):
        self.api_key = api_key
        self.logger = logger
        self.ps_client_name = ps_client_name
        self.client = self._build_client(self.CLIENT_CLASS, api_key, logger=logger)

    def download(self, job_id, destination):
        files = self._get_files_list(job_id)
        s3_downloader = S3FilesDownloader(logger=self.logger)
        s3_downloader.download_list(files, destination)

    @abc.abstractmethod
    def _get_files_list(self, job_id):
        """
        :param str job_id:
        :returns: Tuple of (file path, url) pairs
        :rtype: tuple[tuple[str,str]]
        """
        pass

    def _build_client(self, client_class, *args, **kwargs):
        """
        :param type[BaseClient] client_class:
        """
        client = client_class(*args, **kwargs)

        if self.ps_client_name is not None:
            client.ps_client_name = self.ps_client_name

        return client


class JobArtifactsDownloader(ResourceDownloader):
    CLIENT_CLASS = JobsClient

    def _get_files_list(self, job_id):
        start_after = None
        files = []
        while True:
            pagination_response = self.client.artifacts_list(job_id, start_after=start_after)

            if pagination_response.data:
                files.extend(pagination_response.data)
            start_after = pagination_response.start_after

            if start_after is None:
                break

        files = tuple((f.file, f.url) for f in files)
        return files


class ModelFilesDownloader(ResourceDownloader):
    CLIENT_CLASS = ModelsClient

    def _get_files_list(self, model_id):
        files = self.client.get_model_files(model_id=model_id, links=True)
        files = tuple((f.file, f.url) for f in files)
        return files



import marshmallow

from .base import BaseSchema
from .tag import TagSchema
from .. import models


class MachineEventSchema(BaseSchema):
    MODEL = models.MachineEvent

    name = marshmallow.fields.Str()
    state = marshmallow.fields.Str()
    created = marshmallow.fields.Str(dump_to="dtCreated", load_from="dtCreated")


class MachineSchema(BaseSchema):
    MODEL = models.Machine

    id = marshmallow.fields.Str(dump_to="machineId", load_from="machineId")
    name = marshmallow.fields.Str(dump_to="machineName", load_from="machineName")
    machine_type = marshmallow.fields.Str(dump_to="machineType", load_from="machineType")
    region = marshmallow.fields.Str()
    size = marshmallow.fields.Int()
    billing_type = marshmallow.fields.Str(dump_to="billingType", load_from="billingType")
    template_id = marshmallow.fields.Str(dump_to="templateId", load_from="templateId")
    assign_public_ip = marshmallow.fields.Bool(dump_to="assignPublicIp", load_from="assignPublicIp")
    dynamic_public_ip = marshmallow.fields.Bool(dump_to="dynamicPublicIp", load_from="dynamicPublicIp")
    network_id = marshmallow.fields.Str(dump_to="networkId", load_from="networkId")
    team_id = marshmallow.fields.Str(dump_to="teamId", load_from="teamId")
    user_id = marshmallow.fields.Str(dump_to="userId", load_from="userId")
    email = marshmallow.fields.Str()
    password = marshmallow.fields.Str()
    first_name = marshmallow.fields.Str(dump_to="firstName", load_from="firstName")
    last_name = marshmallow.fields.Str(dump_to="lastName", load_from="lastName")
    notification_email = marshmallow.fields.Str(dump_to="notificationEmail", load_from="notificationEmail")
    script_id = marshmallow.fields.Str(dump_to="scriptId", load_from="scriptId")
    os = marshmallow.fields.Str()
    cpus = marshmallow.fields.Int()
    ram = marshmallow.fields.Int()
    gpu = marshmallow.fields.Str()
    state = marshmallow.fields.Str()

    updates_pending = marshmallow.fields.Bool(dump_to="updatesPending", load_from="updatesPending")
    perform_auto_snapshot = marshmallow.fields.Bool(dump_to="performAutoSnapshot", load_from="performAutoSnapshot")
    auto_snapshot_frequency = marshmallow.fields.Str(dump_to="autoSnapshotFrequency", load_from="autoSnapshotFrequency")
    auto_snapshot_save_count = marshmallow.fields.Int(dump_to="autoSnapshotSaveCount",
                                                      load_from="autoSnapshotSaveCount")
    shutdown_timeout_in_hours = marshmallow.fields.Int(dump_to="shutdownTimeoutInHours",
                                                       load_from="shutdownTimeoutInHours")
    shutdown_timeout_forces = marshmallow.fields.Bool(dump_to="shutdownTimeoutForces",
                                                      load_from="shutdownTimeoutForces")
    agent_type = marshmallow.fields.Str(dump_to="agentType", load_from="agentType")
    storage_total = marshmallow.fields.Str(dump_to="storageTotal", load_from="storageTotal")
    storage_used = marshmallow.fields.Str(dump_to="storageUsed", load_from="storageUsed")
    public_ip_address = marshmallow.fields.Str(dump_to="publicIpAddress", load_from="publicIpAddress")
    private_ip_address = marshmallow.fields.Str(dump_to="privateIpAddress", load_from="privateIpAddress")
    usage_rate = marshmallow.fields.Str(dump_to="usageRate", load_from="usageRate")
    created_timestamp = marshmallow.fields.Str(dump_to="dtCreated", load_from="dtCreated")
    last_run_timestamp = marshmallow.fields.Str(dump_to="dtLastRun", load_from="dtLastRun")

    events = marshmallow.fields.Nested(MachineEventSchema, many=True)
    tags = marshmallow.fields.Nested(TagSchema, only="name", many=True, load_only=True)


class MachineSchemaForListing(MachineSchema):
    name = marshmallow.fields.Str()



import abc

import six

from gradient import api_sdk, exceptions
from gradient.api_sdk import sdk_exceptions
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands.common import BaseCommand, ListCommandMixin, DetailsCommandMixin
from gradient.commands.experiments import BaseCreateExperimentCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseHyperparameterCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.HyperparameterJobsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class CreateHyperparameterCommand(BaseCreateExperimentCommandMixin, BaseHyperparameterCommand):
    SPINNER_MESSAGE = "Creating hyperparameter tuning job"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "Hyperparameter tuning job created with ID: {}"

    def _create(self, hyperparameter):
        handle = self.client.create(**hyperparameter)
        return handle


class CreateAndStartHyperparameterCommand(BaseCreateExperimentCommandMixin, BaseHyperparameterCommand):
    SPINNER_MESSAGE = "Creating and starting hyperparameter tuning job"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "Hyperparameter tuning job created and started with ID: {}"

    def _create(self, hyperparameter):
        handle = self.client.run(**hyperparameter)
        return handle


class ListHyperparametersCommand(ListCommandMixin, BaseHyperparameterCommand):
    def _get_instances(self, kwargs):
        try:
            instances = self.client.list()
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances

    def _get_table_data(self, objects):
        data = [("Name", "ID", "Project ID")]
        for obj in objects:
            data.append((obj.name, obj.id, obj.project_id))
        return data


class HyperparameterDetailsCommand(DetailsCommandMixin, BaseHyperparameterCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def _get_table_data(self, instance):
        """
        :param api_sdk.Hyperparameter instance:
        """
        data = (
            ("ID", instance.id),
            ("Name", instance.name),
            ("Ports", instance.ports),
            ("Project ID", instance.project_id),
            ("Tuning command", instance.tuning_command),
            ("Worker command", instance.worker_command),
            ("Worker container", instance.worker_container),
            ("Worker count", instance.worker_count),
            ("Worker machine type", instance.worker_machine_type),
            ("Worker use dockerfile", instance.use_dockerfile or False),
            ("Workspace URL", instance.workspace_url),
        )
        return data


class HyperparameterStartCommand(BaseHyperparameterCommand):
    def execute(self, id_):
        self.client.start(id_)
        self.logger.log("Hyperparameter tuning started")


class HyperparameterAddTagsCommand(BaseHyperparameterCommand):
    def execute(self, hyperparameter_id, *args, **kwargs):
        self.client.add_tags(hyperparameter_id, **kwargs)
        self.logger.log("Tags added to hyperparameter")


class HyperparameterRemoveTagsCommand(BaseHyperparameterCommand):
    def execute(self, hyperparameter_id, *args, **kwargs):
        self.client.remove_tags(hyperparameter_id, **kwargs)
        self.logger.log("Tags removed from hyperparameter")



"""https://stackoverflow.com/questions/40753999/python-click-make-option-value-optional"""

import click
from click_didyoumean import DYMMixin
from click_help_colors import HelpColorsMixin

from gradient.cli.common import GradientOption


class RegisterReaderOptionMixin(object):
    """ Mark this option as getting a _set option """
    register_reader = True


class RegisterWriterOptionMixin(object):
    """ Fix the help for the _set suffix """

    def get_help_record(self, ctx):
        help = super(RegisterWriterOptionMixin, self).get_help_record(ctx)
        return (help[0].replace('_set ', '='),) + help[1:]


class RegisterWriterCommandMixin(object):
    def parse_args(self, ctx, args):
        """ Translate any opt= to opt_set= as needed """
        options = [o for o in ctx.command.params
                   if getattr(o, 'register_reader', None)]
        prefixes = {p for p in sum([o.opts for o in options], [])
                    if p.startswith('--')}
        for i, a in enumerate(args):
            try:
                a = a.split('=')
            except AttributeError:
                continue  # skip the following if a is not a string

            if a[0] in prefixes and len(a) > 1:
                a[0] += '_set'
                args[i] = '='.join(a)

        return super(RegisterWriterCommandMixin, self).parse_args(ctx, args)


class GradientRegisterReaderOption(RegisterReaderOptionMixin, GradientOption):
    pass


class GradientRegisterWriterOption(RegisterWriterOptionMixin, GradientOption):
    pass


class GradientRegisterWriterCommand(DYMMixin, HelpColorsMixin, RegisterWriterCommandMixin, click.Command):
    pass



import click
import pytest

from gradient.cli.deployments import validate_autoscaling_metric_or_resource


def test_should_return_none_if_none_was_provided():
    rv = validate_autoscaling_metric_or_resource(1, 2, None, "Metric")
    assert rv is None


def test_should_raise_bad_parameter_if_provided_value_was_not_valid():
    with pytest.raises(click.BadParameter):
        validate_autoscaling_metric_or_resource(1, 2, ["cpu"], "Metric")


def test_should_return_valid_structure_if_value_format_was_correct():
    expected_value = (
        {
            "type": "Metric",
            "name": "cpu",
            "value_type": "target",
            "value": 60,
        },
        {
            "type": "Metric",
            "name": "gpu",
            "value_type": "target",
            "value": 40,
        },
    )

    rv = validate_autoscaling_metric_or_resource(1, 2, ["cpu/target:60", "gpu/target:40"], "Metric")

    assert rv == expected_value



class ApplicationError(Exception):
    pass


class BadResponseError(ApplicationError):
    pass


class PresignedUrlUnreachableError(ApplicationError):
    pass


class ProjectAccessDeniedError(ApplicationError):
    pass


class PresignedUrlAccessDeniedError(ApplicationError):
    pass


class PresignedUrlConnectionError(ApplicationError):
    pass


class PresignedUrlMalformedResponseError(ApplicationError):
    pass


class PresignedUrlError(ApplicationError):
    pass


class S3UploadFailedError(ApplicationError):
    pass


class ReceivingDataFailedError(ApplicationError):
    pass


class WrongPathError(ApplicationError):
    pass


class MutuallyExclusiveParametersUsedError(Exception):
    pass



import datetime

import attr

from .. import constants


@attr.s
class BaseExperiment(object):
    """
    Base experiment class. Single node and multi node experiments classes inherit from it.

    Experiments are intended to be used for intensive computational tasks like neural network training. Gradient
    supports single-node experiments as well as distributed training through multinode experiments.

    Experiments can be run from the Experiment Builder web interface, the GradientCI bot, or the CLI.

    :param str name: Name of new experiment
    :param str ports: Port to use in new experiment
    :param str workspace_url: Project git repository url
    :param str workspace_ref: Git commit hash, branch name or tag
    :param str workspace_username: Project git repository username
    :param str workspace_password: Project git repository password
    :param Dataset dataset
    :param str working_directory: Working directory for the experiment
    :param str artifact_directory: Artifacts directory
    :param str cluster_id: Cluster ID (handle)
    :param dict experiment_env: Environment variables in a JSON
    :param str project_id: Project ID
    :param str model_type: defines the type of model that is being generated by the experiment.
    :param str model_path: Model path
    :param bool is_preemptible: Is preemptible
    :param str id:
    :param int state: state of experiment can be one of::

            "created"
            "provisioned"
            "network setup"
            "running"
            "stopped"
            "error"
            "failed"
            "canceled"
            "network teardown"
            "pending"
            "provisioning"
            "network setting up"
            "network tearing down"
    """
    name = attr.ib(type=str, default=None)
    ports = attr.ib(type=str, default=None)
    workspace_url = attr.ib(type=str, default=None)
    workspace_ref = attr.ib(type=str, default=None)
    workspace_username = attr.ib(type=str, default=None)
    workspace_password = attr.ib(type=str, default=None)
    datasets = attr.ib(type=list, default=None)
    working_directory = attr.ib(type=str, default=None)
    artifact_directory = attr.ib(type=str, default=None)
    cluster_id = attr.ib(type=str, default=None)
    experiment_env = attr.ib(type=dict, default=dict)
    project_id = attr.ib(type=str, default=None)
    model_type = attr.ib(type=str, default=None)
    model_path = attr.ib(type=str, default=None)
    is_preemptible = attr.ib(type=bool, default=None)
    id = attr.ib(type=str, default=None)
    state = attr.ib(type=int, default=None)
    tags = attr.ib(type=list, factory=list)

    dt_created = attr.ib(type=datetime.datetime, default=None)
    dt_modified = attr.ib(type=datetime.datetime, default=None)
    dt_started = attr.ib(type=datetime.datetime, default=None)
    dt_stopped = attr.ib(type=datetime.datetime, default=None)
    dt_deleted = attr.ib(type=datetime.datetime, default=None)


@attr.s
class SingleNodeExperiment(BaseExperiment):
    """
    Single node experiment class. Inherits from ``BaseExperiment`` class

    In your your CLI command or ``config.yaml``, specify the experiment type as ``singlenode``

    :param str container: Container (dockerfile)
    :param str machine_type: Machine type
        Options::

        "Air"
        "Standard"
        "Pro"
        "Advanced"
        "GPU+"
        "P4000"
        "P5000"
        "P6000"
        "V100"
        "C1"
        "C2"
        "C3"
        "C4"
        "C5"
        "C6"
        "C7"
        "C8"
        "C9"
        "C10"
    :param str command: Container entrypoint command
    :param str container_user: Container user for running the specified command in the container. If no containerUser is specified, the user will default to 'root' in the container.
    :param str registry_username: Registry username for accessing private docker registry container if nessesary
    :param str registry_password: Registry password for accessing private docker registry container if nessesary
    :param int experiment_type_id: type of experiment
        Options::

            "single node"
            "GRPC multi node"
            "MPI multi node"
            "Hyperparameter tuning"
    """
    container = attr.ib(type=str, default=None)
    machine_type = attr.ib(type=str, default=None)
    command = attr.ib(type=str, default=None)
    container_user = attr.ib(type=str, default=None)
    registry_username = attr.ib(type=str, default=None)
    registry_password = attr.ib(type=str, default=None)
    registry_url = attr.ib(type=str, default=None)
    experiment_type_id = attr.ib(type=int, default=constants.ExperimentType.SINGLE_NODE)

    @experiment_type_id.validator
    def experiment_type_id_validator(self, attribute, value):
        if value is not constants.ExperimentType.SINGLE_NODE:
            raise ValueError("Single node experiment's type must equal {}".
                             format(constants.ExperimentType.SINGLE_NODE))


@attr.s
class MultiNodeExperiment(BaseExperiment):
    """
    Multi node experiment class. Inherits from ``BaseExperiment`` class

    Gradient supports both gRPC and MPI protocols for distributed TensorFlow model training. In your CLI command or
    ``config.yaml``, specify the experiment type as either ``multinode``.

    The two types are::

        type: "multi-grpc"

    or::

        type: "multi-mpi"

    :param int experiment_type_id: type of experiment
        Options::

            "single node"
            "GRPC multi node"
            "MPI multi node"
            "Hyperparameter tuning"
    :param str worker_container: Worker container (dockerfile)
    :param str worker_machine_type: Worker machine type
    :param str worker_command: Worker command
    :param int worker_count: Worker count
    :param str parameter_server_container: Parameter server container
    :param str parameter_server_command: Parameter server command
    :param int parameter_server_count: Parameter server count
    :param str worker_container_user: Worker container user
    :param str worker_registry_username: Registry username for accessing private docker registry container if nessesary
    :param str worker_registry_password: Registry password for accessing private docker registry container if nessesary
    :param str parameter_server_container_user: Parameter server container user
    :param str parameter_server_registry_username: Registry username for accessing private docker registry container if nessesary
    :param str parameter_server_registry_password: Registry password for accessing private docker registry container if nessesary
    """
    experiment_type_id = attr.ib(type=int, default=None)
    worker_container = attr.ib(type=str, default=None)
    worker_machine_type = attr.ib(type=str, default=None)
    worker_command = attr.ib(type=str, default=None)
    worker_count = attr.ib(type=int, default=None)
    parameter_server_container = attr.ib(type=str, default=None)
    parameter_server_machine_type = attr.ib(type=str, default=None)
    parameter_server_command = attr.ib(type=str, default=None)
    parameter_server_count = attr.ib(type=int, default=None)
    worker_container_user = attr.ib(type=str, default=None)
    worker_registry_username = attr.ib(type=str, default=None)
    worker_registry_password = attr.ib(type=str, default=None)
    worker_registry_url = attr.ib(type=str, default=None)
    parameter_server_container_user = attr.ib(type=str, default=None)
    parameter_server_registry_username = attr.ib(type=str, default=None)
    parameter_server_registry_password = attr.ib(type=str, default=None)
    parameter_server_registry_url = attr.ib(type=str, default=None)

    @experiment_type_id.validator
    def experiment_type_id_validator(self, attribute, value):
        if value not in (constants.ExperimentType.GRPC_MULTI_NODE,
                         constants.ExperimentType.MPI_MULTI_NODE):
            raise ValueError("Multi node experiment's type must equal {} or {}".
                             format(constants.ExperimentType.GRPC_MULTI_NODE,
                                    constants.ExperimentType.MPI_MULTI_NODE))


@attr.s
class MpiMultiNodeExperiment(BaseExperiment):
    experiment_type_id = attr.ib(type=int, default=None)
    worker_container = attr.ib(type=str, default=None)
    worker_machine_type = attr.ib(type=str, default=None)
    worker_command = attr.ib(type=str, default=None)
    worker_count = attr.ib(type=int, default=None)
    master_container = attr.ib(type=str, default=None)
    master_machine_type = attr.ib(type=str, default=None)
    master_command = attr.ib(type=str, default=None)
    master_count = attr.ib(type=str, default=None)
    worker_container_user = attr.ib(type=str, default=None)
    worker_registry_username = attr.ib(type=str, default=None)
    worker_registry_password = attr.ib(type=str, default=None)
    worker_registry_url = attr.ib(type=str, default=None)
    master_container_user = attr.ib(type=str, default=None)
    master_registry_username = attr.ib(type=str, default=None)
    master_registry_password = attr.ib(type=str, default=None)
    master_registry_url = attr.ib(type=str, default=None)



import os

import mock
import pytest

import gradient.api_sdk.utils
import gradient.cliutils
from gradient.api_sdk.workspace import S3WorkspaceHandler
from gradient.cli_constants import CLI_PS_CLIENT_NAME

MOCK_BUCKET_NAME = 'bucket_name'
MOCK_OBJECT_KEY = 'object_key'
mock_upload_data = {
    "bucket_name": MOCK_BUCKET_NAME,
    "fields": {
        "key": MOCK_OBJECT_KEY
    }
}

mock_upload_response = {
    "message": "success",
    "data": mock_upload_data
}


@pytest.fixture
def workspace_handler():
    s3_workspace_handler = S3WorkspaceHandler("some_key", client_name=CLI_PS_CLIENT_NAME)
    s3_workspace_handler._upload = mock.MagicMock(return_value="s3://{}/{}".format(MOCK_BUCKET_NAME, MOCK_OBJECT_KEY))
    return s3_workspace_handler


class TestWorkspace(object):
    @mock.patch("gradient.utils.PathParser.parse_path", return_value=gradient.api_sdk.utils.PathParser.S3_URL)
    @mock.patch("gradient.api_sdk.workspace.S3WorkspaceHandler._upload")
    def test_dont_upload_if_s3_url_provided(self, _, __, workspace_handler):
        workspace_handler._upload = mock.MagicMock()

        workspace_handler.handle({'workspace': 's3://some-path'})

        workspace_handler._upload.assert_not_called()

    def test_zip_files_and_receive_s3_response_when_no_dir_provided(self, workspace_handler):
        archive_name = 'foo.zip'

        workspace_handler._zip_workspace = mock.MagicMock()
        workspace_handler._zip_workspace.return_value = archive_name

        response_url = workspace_handler.handle({"projectHandle": "some_project_id"})

        workspace_handler._zip_workspace.assert_not_called()
        workspace_handler._upload.assert_not_called()
        assert response_url is None

    @mock.patch("gradient.utils.PathParser.is_local_dir", return_value=True)
    def test_zip_files_and_receive_s3_response_when_workspace_dir_provided(self, _, workspace_handler):
        archive_name = 'foo.zip'

        workspace_handler._zip_workspace = mock.MagicMock()
        workspace_handler._zip_workspace.return_value = archive_name

        response_url = workspace_handler.handle({"projectHandle": "some_project_id", "workspace": "foo/bar"})

        workspace_handler._zip_workspace.assert_called_once()
        workspace_handler._upload.assert_called_once()
        workspace_handler._upload.assert_called_with(archive_name, "some_project_id", cluster_id=None)
        assert response_url == 's3://{}/{}'.format(MOCK_BUCKET_NAME, MOCK_OBJECT_KEY)

    @mock.patch("gradient.utils.PathParser.is_local_zip_file", return_value=True)
    def test_dont_zip_files_and_receive_s3_response_when_workspace_archive_provided(self, _, workspace_handler):
        workspace_handler._zip_workspace = mock.MagicMock()

        response_url = workspace_handler.handle({"projectHandle": "some_project_id", "workspace": "foo.zip"})

        workspace_handler._zip_workspace.assert_not_called()
        workspace_handler._upload.assert_called_once()
        workspace_handler._upload.assert_called_with(os.path.abspath('foo.zip'), "some_project_id", cluster_id=None)
        assert response_url == 's3://{}/{}'.format(MOCK_BUCKET_NAME, MOCK_OBJECT_KEY)

    @mock.patch("gradient.utils.PathParser.is_local_zip_file", return_value=True)
    def test_dont_zip_files_and_receive_s3_response_when_workspace_archive_provided_with_workspace(self, _,
                                                                                                   workspace_handler):
        workspace_handler._zip_workspace = mock.MagicMock()

        response_url = workspace_handler.handle({'projectHandle': 'some_project_id', 'workspace': 'foo.zip'})

        workspace_handler._zip_workspace.assert_not_called()
        workspace_handler._upload.assert_called_once()
        workspace_handler._upload.assert_called_with(os.path.abspath('foo.zip'), "some_project_id", cluster_id=None)
        assert response_url == 's3://{}/{}'.format(MOCK_BUCKET_NAME, MOCK_OBJECT_KEY)



"""
Deployment related client handler logic.

Remember that in code snippets all highlighted lines are required other lines are optional.
"""
from .base_client import BaseClient, TagsSupportMixin
from .. import config, models, repositories


class DeploymentsClient(TagsSupportMixin, BaseClient):
    """
    Client to handle deployment related actions.

    How to create instance of deployment client:

    .. code-block:: python
        :linenos:
        :emphasize-lines: 4

        from gradient import DeploymentsClient

        deployment_client = DeploymentsClient(
            api_key='your_api_key_here'
        )
    """
    HOST_URL = config.config.CONFIG_HOST
    entity = "deployment"

    def create(
            self,
            deployment_type,
            name,
            machine_type,
            image_url,
            instance_count=None,
            model_id=None,
            container_model_path=None,
            image_username=None,
            image_password=None,
            image_server=None,
            container_url_path=None,
            endpoint_url_path=None,
            method=None,
            docker_args=None,
            env=None,
            api_type=None,
            ports=None,
            cluster_id=None,
            auth_username=None,
            auth_password=None,
            tags=None,
            command=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            project_id=None,
            autoscaling=None,
    ):
        """
        Method to create a Deployment instance.

        To create a new Deployment, you must first create a Model. With a Model available, use the ``create`` subcommand
        and specify all of the following parameters: deployment type, base image, name, machine type, and container
        image for serving, as well as the instance count:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 4

            from gradient import DeploymentsClient

            deployment_client = DeploymentsClient(
                api_key='your_api_key_here'
            )


        To obtain your Model ID, you can run ``command gradient models list`` and copy the target Model ID from
        your available Models.

        :param str deployment_type: Model deployment type. Only TensorFlow Model deployment type is currently supported  [required]
        :param str model_id: ID of a trained model [required]
        :param str name: Human-friendly name for new model deployment [required]
        :param str machine_type: [G1|G6|G12|K80|P100|GV100] Type of machine for new deployment [required]
        :param str image_url: Docker image for model deployment  [required]
        :param int instance_count: Number of machine instances  [required]
        :param str container_model_path: Container model path
        :param str image_username: Username used to access docker image
        :param str image_password: Password used to access docker image
        :param str image_server: Docker image server
        :param str container_url_path: Container URL path
        :param str endpoint_url_path: Endpoint URL path
        :param str method: Method
        :param list[str]|tuple[str] docker_args: List of docker args
        :param dict[str,str] env: Environmental variables
        :param str api_type: Type of API (REST/GRPC)
        :param str ports: Ports
        :param str cluster_id: cluster ID
        :param str auth_username: Username
        :param str auth_password: Password
        :param list[str] tags: List of tags
        :param str command: Deployment command
        :param str workspace_url: Project git or s3repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param str project_id: Project ID
        :param models.AutoscalingDefinition autoscaling: Deployment autoscaling definition

        :returns: Created deployment id
        :rtype: str
        """
        deployment = models.Deployment(
            deployment_type=deployment_type,
            model_id=model_id,
            name=name,
            machine_type=machine_type,
            image_url=image_url,
            instance_count=instance_count,
            container_model_path=container_model_path,
            image_username=image_username,
            image_password=image_password,
            image_server=image_server,
            container_url_path=container_url_path,
            endpoint_url_path=endpoint_url_path,
            method=method,
            docker_args=docker_args,
            env=env,
            api_type=api_type,
            ports=ports,
            cluster_id=cluster_id,
            auth_username=auth_username,
            auth_password=auth_password,
            command=command,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            project_id=project_id,
            autoscaling=autoscaling,
        )

        repository = self.build_repository(repositories.CreateDeployment)
        deployment_id = repository.create(deployment)
        if tags:
            self.add_tags(entity_id=deployment_id, tags=tags)
        return deployment_id

    def get(self, deployment_id):
        """Get deployment instance

        :param str deployment_id: Deployment ID

        :return: Deployment instance
        :rtype: models.Deployment
        """
        repository = self.build_repository(repositories.GetDeployment)
        deployment = repository.get(deployment_id=deployment_id)
        return deployment

    def start(self, deployment_id):
        """
        Start deployment

        *EXAMPLE*::

            gradient deployments start --id <your-deployment-id>

        :param str deployment_id: Deployment ID
        """

        repository = self.build_repository(repositories.StartDeployment)
        repository.start(deployment_id)

    def stop(self, deployment_id):
        """
        Stop deployment

        *EXAMPLE*::

            gradient deployments stop --id <your-deployment-id>

        :param deployment_id: Deployment ID
        """

        repository = self.build_repository(repositories.StopDeployment)
        repository.stop(deployment_id)

    def list(self, state=None, project_id=None, model_id=None, tags=None):
        """
        List deployments with optional filtering

        :param str state: state to filter deployments
        :param str project_id: project ID to filter deployments
        :param str model_id: model ID to filter deployments
        :param list[str]|tuple[str] tags: tags to filter deployments with OR

        :returns: List of Deployment model instances
        :rtype: list[models.Deployment]
        """

        repository = self.build_repository(repositories.ListDeployments)
        deployments = repository.list(state=state, project_id=project_id, model_id=model_id, tags=tags)
        return deployments

    def delete(self, deployment_id):
        repository = self.build_repository(repositories.DeleteDeployment)
        repository.delete(deployment_id)

    def update(
            self,
            deployment_id,
            deployment_type=None,
            model_id=None,
            name=None,
            machine_type=None,
            image_url=None,
            instance_count=None,
            container_model_path=None,
            image_username=None,
            image_password=None,
            image_server=None,
            container_url_path=None,
            endpoint_url_path=None,
            method=None,
            docker_args=None,
            env=None,
            api_type=None,
            ports=None,
            cluster_id=None,
            auth_username=None,
            auth_password=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            project_id=None,
            command=None,
            autoscaling=None,
    ):
        deployment = models.Deployment(
            deployment_type=deployment_type,
            model_id=model_id,
            name=name,
            machine_type=machine_type,
            image_url=image_url,
            instance_count=instance_count,
            container_model_path=container_model_path,
            image_username=image_username,
            image_password=image_password,
            image_server=image_server,
            container_url_path=container_url_path,
            endpoint_url_path=endpoint_url_path,
            method=method,
            docker_args=docker_args,
            env=env,
            api_type=api_type,
            ports=ports,
            cluster_id=cluster_id,
            auth_username=auth_username,
            auth_password=auth_password,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            project_id=project_id,
            command=command,
            autoscaling=autoscaling,
        )

        repository = self.build_repository(repositories.UpdateDeployment)
        repository.update(deployment_id, deployment)

    def get_metrics(self, deployment_id, start=None, end=None, interval="30s", built_in_metrics=None):
        """Get model deployment metrics

        :param str deployment_id: ID of deployment
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Metrics of a model deployment job
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.GetDeploymentMetrics)
        metrics = repository.get(
            id=deployment_id,
            start=start,
            end=end,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    
    def list_metrics(self, deployment_id, start=None, end=None, interval="30s"):
        """List model deployment metrics

        :param str deployment_id: ID of deployment
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :returns: Metrics of a model deployment job
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.ListDeploymentMetrics)
        metrics = repository.get(
            id=deployment_id,
            start=start,
            end=end,
            interval=interval,
        )
        return metrics

    def stream_metrics(self, deployment_id, interval="30s", built_in_metrics=None):
        """Stream live model deployment metrics

        :param str deployment_id: ID of model deployment
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Generator object yielding live model deployment metrics
        :rtype: Iterable[dict]
        """

        repository = self.build_repository(repositories.StreamDeploymentMetrics)
        metrics = repository.stream(
            id=deployment_id,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def logs(self, deployment_id, line=1, limit=10000):
        """Show list of latest logs from the specified deployment.

        :param str deployment_id: Deployment Id
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: list of LogRows
        :rtype: list[models.LogRow]
        """

        repository = self.build_repository(repositories.ListDeploymentLogs)
        logs = repository.list(id=deployment_id, line=line, limit=limit)
        return logs

    def yield_logs(self, deployment_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        :param str deployment_id: Deployment Id
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        repository = self.build_repository(repositories.ListDeploymentLogs)
        logs = repository.yield_logs(id=deployment_id, line=line, limit=limit)
        return logs



import mock
from click.testing import CliRunner

from gradient.api_sdk.clients import http_client
from gradient.cli import cli
from tests import example_responses, MockResponse

EXPECTED_HEADERS = http_client.default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"

URL = "https://api.paperspace.io"


class TestListDatasets(object):
    COMMAND = ["datasets", "list"]

    JSON = example_responses.LIST_DATASETS_RESPONSE

    STDOUT = """+-------+-----------------+-------------------------+
| Name  | ID              | Storage Provider        |
+-------+-----------------+-------------------------+
| test1 | dsttn2y7j1ux882 | test1 (spltautet072md4) |
+-------+-----------------+-------------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_list_datasets(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND)

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets",
            headers=EXPECTED_HEADERS,
            json=None,
            params={
                "filter[limit]": 21,
                "filter[skip]": 0,
                "filter[order][]": "name ASC",
            }
        )


class TestShowDatasetDetails(object):
    COMMAND = ["datasets", "details"]

    JSON = example_responses.SHOW_DATASET_DETAILS_RESPONSE

    STDOUT = """+-----------------+-------------------------+
| Name            | test1                   |
+-----------------+-------------------------+
| ID              | dsttn2y7j1ux882         |
| Description     |                         |
| StorageProvider | test1 (spltautet072md4) |
+-----------------+-------------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_show_dataset_details(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=dsttn2y7j1ux882"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882",
            headers=EXPECTED_HEADERS,
            json=None,
            params=None
        )


class TestCreateDataset(object):
    COMMAND = ["datasets", "create"]

    JSON = example_responses.CREATE_DATASET_RESPONSE

    STDOUT = """Created dataset: dsttn2y7j1ux882
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_create_dataset(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "--name=test1", "--storageProviderId=spltautet072md4"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets",
            json={"name": "test1", "storageProviderId": "spltautet072md4"},
            params=None,
            headers=EXPECTED_HEADERS,
            files=None,
            data=None
        )


class TestUpdateDataset(object):
    COMMAND = ["datasets", "update"]

    JSON = example_responses.UPDATE_DATASET_RESPONSE

    STDOUT = """Updated dataset: dsttn2y7j1ux882
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_update_dataset(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "--id=dsttn2y7j1ux882", "--description=Test dataset"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882",
            json={"description": "Test dataset"},
            params=None,
            headers=EXPECTED_HEADERS,
            files=None,
            data=None
        )


class TestDeleteDataset(object):
    COMMAND = ["datasets", "delete"]

    STDOUT = """Deleted dataset: dsttn2y7j1ux882
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_delete_dataset(self, method):
        method.return_value = MockResponse(status_code=204)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=dsttn2y7j1ux882"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882",
            params=None,
            headers=EXPECTED_HEADERS,
            json=None
        )


class TestListDatasetVersions(object):
    COMMAND = ["datasets", "versions", "list"]

    JSON = example_responses.LIST_DATASET_VERSIONS_RESPONSE

    STDOUT = """+-------------------------+---------+-------+
| ID                      | Message | Tags  |
+-------------------------+---------+-------+
| dsttn2y7j1ux882:1rn19s2 |         | hello |
+-------------------------+---------+-------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_list_dataset_versions(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=dsttn2y7j1ux882"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/versions",
            headers=EXPECTED_HEADERS,
            json=None,
            params={
                "filter[limit]": 21,
                "filter[skip]": 0,
                "filter[where][isCommitted]": "true",
                "filter[order][]": "dtCreated DESC",
            }
        )


class TestShowDatasetVersionDetails(object):
    COMMAND = ["datasets", "versions", "details"]

    JSON = example_responses.SHOW_DATASET_VERSION_DETAILS_RESPONSE

    STDOUT = """+-----------+-------------------------+
| ID        | dsttn2y7j1ux882:1rn19s2 |
+-----------+-------------------------+
| Message   |                         |
| Committed | true                    |
| Tags      | hello                   |
+-----------+-------------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_show_dataset_details(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=dsttn2y7j1ux882:1rn19s2"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/versions/1rn19s2",
            headers=EXPECTED_HEADERS,
            json=None,
            params=None
        )


class TestUpdateDatasetVersion(object):
    COMMAND = ["datasets", "versions", "update"]

    JSON = example_responses.UPDATE_DATASET_VERSION_RESPONSE

    STDOUT = """Updated dataset version: dsttn2y7j1ux882:1rn19s2
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_update_dataset_version(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "--id=dsttn2y7j1ux882:1rn19s2", "--message=Test message"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/versions/1rn19s2",
            json={"message": "Test message"},
            params=None,
            headers=EXPECTED_HEADERS,
            files=None,
            data=None
        )


class TestDeleteDatasetVersion(object):
    COMMAND = ["datasets", "versions", "delete"]

    STDOUT = """Deleted dataset version: dsttn2y7j1ux882:1rn19s2
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_delete_dataset_version(self, method):
        method.return_value = MockResponse(status_code=204)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=dsttn2y7j1ux882:1rn19s2"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/versions/1rn19s2",
            params=None,
            headers=EXPECTED_HEADERS,
            json=None
        )


class TestSetDatasetTag(object):
    COMMAND = ["datasets", "tags", "set"]

    JSON = example_responses.SET_DATASET_VERSION_TAG_RESPONSE

    STDOUT = """Set dataset tag: dsttn2y7j1ux882:hello
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_update_dataset_version(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "--id=dsttn2y7j1ux882:1rn19s2", "--name=hello"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/tags/hello",
            json={"version": "1rn19s2"},
            params=None,
            headers=EXPECTED_HEADERS,
            data=None
        )


class TestDeleteDatasetTag(object):
    COMMAND = ["datasets", "tags", "delete"]

    STDOUT = """Deleted dataset tag: dsttn2y7j1ux882:hello
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_update_dataset_version(self, method):
        method.return_value = MockResponse()

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "--id=dsttn2y7j1ux882:hello"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/datasets/dsttn2y7j1ux882/tags/hello",
            params=None,
            headers=EXPECTED_HEADERS,
            json=None
        )



import click

from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.common import validate_comma_split_option
from gradient.commands import projects as projects_commands


@cli.group("projects", help="Manage projects", cls=common.ClickGroup)
def projects_group():
    pass


@projects_group.group("tags", help="Manage project tags", cls=common.ClickGroup)
def project_tags():
    pass


@projects_group.command("list", help="List projects")
@click.option(
    "--tag",
    "tags",
    multiple=True,
    cls=common.GradientOption,
    help="Filter by tags. Multiple use"
)
@common.api_key_option
@common.options_file
def list_projects(api_key, tags, options_file):
    command = projects_commands.ListProjectsCommand(api_key=api_key)
    command.execute(tags=tags)


@projects_group.command("create", help="Create project")
@click.option(
    "--name",
    "name",
    required=True,
    help="Name of new project",
    cls=common.GradientOption,
)
@click.option(
    "--repositoryName",
    "repository_name",
    help="Name of the repository",
    cls=common.GradientOption,
)
@click.option(
    "--repositoryUrl",
    "repository_url",
    help="URL to the repository",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to experiment",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def create_project(api_key, options_file, **project):
    project["tags"] = validate_comma_split_option(project.pop("tags_comma"), project.pop("tags"))
    command = projects_commands.CreateProjectCommand(api_key)
    command.execute(project)


@projects_group.command("details", help="Show details of a project")
@click.option(
    "--id",
    "project_id",
    required=True,
    help="Project ID",
)
@common.api_key_option
@common.options_file
def create_project(project_id, api_key, options_file):
    command = projects_commands.ShowProjectDetailsCommand(api_key)
    command.execute(project_id)


@projects_group.command("delete", help="Delete project and all its experiments")
@click.option(
    "--id",
    "project_id",
    required=True,
    help="ID of project",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def delete_project(project_id, options_file, api_key):
    command = projects_commands.DeleteProjectCommand(api_key)
    command.execute(project_id)


@project_tags.command("add", help="Add tags to project")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the project",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to project",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to project",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def project_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = projects_commands.ProjectAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@project_tags.command("remove", help="Remove tags from project")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the project",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from project",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from project",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def project_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = projects_commands.ProjectRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)



import collections

import click

from gradient import cliutils
from gradient import exceptions, clilogger, DEPLOYMENT_TYPES_MAP
from gradient.api_sdk import constants, workspace
from gradient.api_sdk.s3_uploader import DeploymentWorkspaceDirectoryUploader
from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.cli_types import ChoiceType, json_string
from gradient.cli.common import api_key_option, del_if_value_is_none, ClickGroup, validate_comma_split_option
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands import deployments as deployments_commands
from gradient.commands.deployments import DeploymentRemoveTagsCommand, DeploymentAddTagsCommand, \
    GetDeploymentMetricsCommand, ListDeploymentMetricsCommand, StreamDeploymentMetricsCommand, DeploymentLogsCommand


def get_workspace_handler(api_key):
    logger_ = clilogger.CliLogger()
    workspace_handler = workspace.S3WorkspaceHandlerWithProgressbar(api_key=api_key,
                                                                    logger_=logger_,
                                                                    uploader_cls=DeploymentWorkspaceDirectoryUploader,
                                                                    client_name=CLI_PS_CLIENT_NAME)
    return workspace_handler


def validate_autoscaling_metric_or_resource(ctx, param, value, metric_type):
    """
    value in = ("cpu/targetAverage:10")
    value out = ({"type": instance,
                  "name": "cpu",
                  "value_type": "targetAverage",
                  "value": 10})
    """

    if value is None:
        return None

    old_values = value
    new_values = []

    for old_value in old_values:
        try:
            name, values = old_value.split("/", 1)
            value_type, value = values.split(":", 1)
            value = float(value)
        except Exception as e:
            debug_msg = "Error occurred while validating autoscaling {} with value {}: {}" \
                .format(metric_type, old_value, e)
            clilogger.CliLogger().debug(debug_msg)

            msg = "value need to be in format resource_name/value_type:value for example cpu/targetAverage:60" \
                .format(old_value)
            raise click.BadParameter(msg)

        new_value = {"type": metric_type,
                     "name": name,
                     "value_type": value_type,
                     "value": value}
        new_values.append(new_value)

    return tuple(new_values)


def validate_autoscaling_metric(ctx, param, value):
    return validate_autoscaling_metric_or_resource(ctx, param, value, "Metric")


def validate_autoscaling_resource(ctx, param, value):
    return validate_autoscaling_metric_or_resource(ctx, param, value, "Resource")


@cli.group("deployments", help="Manage deployments", cls=ClickGroup)
def deployments_group():
    pass


@deployments_group.group("tags", help="Manage deployments tags", cls=ClickGroup)
def deployments_tags():
    pass


@deployments_group.group(name="metrics", help="Read model deployment metrics", cls=ClickGroup)
def deployments_metrics():
    pass


@deployments_group.command("create", help="Create new deployment")
@click.option(
    "--deploymentType",
    "deployment_type",
    type=ChoiceType(DEPLOYMENT_TYPES_MAP, case_sensitive=False),
    required=True,
    help="Model deployment type",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="Project ID",
    cls=common.GradientOption,
)
@click.option(
    "--modelId",
    "model_id",
    help="ID of a trained model",
    cls=common.GradientOption,
)
@click.option(
    "--name",
    "name",
    required=True,
    help="Human-friendly name for new model deployment",
    cls=common.GradientOption,
)
@click.option(
    "--machineType",
    "machine_type",
    required=True,
    help="Type of machine for new deployment",
    cls=common.GradientOption,
)
@click.option(
    "--imageUrl",
    "image_url",
    required=True,
    help="Docker image for model serving",
    cls=common.GradientOption,
)
@click.option(
    "--instanceCount",
    "instance_count",
    type=int,
    help="Number of machine instances",
    cls=common.GradientOption,
)
@click.option(
    "--command",
    "command",
    help="Deployment command",
    cls=common.GradientOption,
)
@click.option(
    "--containerModelPath",
    "container_model_path",
    help="Container model path",
    cls=common.GradientOption,
)
@click.option(
    "--imageUsername",
    "image_username",
    help="Username used to access docker image",
    cls=common.GradientOption,
)
@click.option(
    "--imagePassword",
    "image_password",
    help="Password used to access docker image",
    cls=common.GradientOption,
)
@click.option(
    "--imageServer",
    "image_server",
    help="Docker image server",
    cls=common.GradientOption,
)
@click.option(
    "--containerUrlPath",
    "container_url_path",
    help="Container URL path",
    cls=common.GradientOption,
)
@click.option(
    "--method",
    "method",
    help="Method",
    cls=common.GradientOption,
)
@click.option(
    "--dockerArgs",
    "docker_args",
    type=json_string,
    help="JSON-style list of docker args",
    cls=common.GradientOption,
)
@click.option(
    "--env",
    "env",
    type=json_string,
    help="JSON-style environmental variables map",
    cls=common.GradientOption,
)
@click.option(
    "--apiType",
    "api_type",
    help="Type of API",
    cls=common.GradientOption,
)
@click.option(
    "--ports",
    "ports",
    help="Ports",
    cls=common.GradientOption,
)
@click.option(
    "--clusterId",
    "cluster_id",
    required=True,
    help="Cluster ID",
    cls=common.GradientOption,
)
@click.option(
    "--authUsername",
    "auth_username",
    help="Username",
    cls=common.GradientOption,
)
@click.option(
    "--authPassword",
    "auth_password",
    help="Password",
    cls=common.GradientOption,
)
@click.option(
    "--auth",
    "generate_auth",
    is_flag=True,
    help="Generate username and password. Mutually exclusive with --authUsername and --authPassword",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to model deployment job",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to model deployment job",
    cls=common.GradientOption
)
@click.option(
    "--workspace",
    "workspace",
    help="Path to workspace directory, archive, S3 or git repository",
    cls=common.GradientOption,
)
@click.option(
    "--workspaceRef",
    "workspace_ref",
    help="Git commit hash, branch name or tag",
    cls=common.GradientOption,
)
@click.option(
    "--workspaceUsername",
    "workspace_username",
    help="Workspace username",
    cls=common.GradientOption,
)
@click.option(
    "--workspacePassword",
    "workspace_password",
    help="Workspace password",
    cls=common.GradientOption,
)
@click.option(
    "--minInstanceCount",
    "min_instance_count",
    help="Minimal instance count",
    cls=common.GradientOption,
)
@click.option(
    "--maxInstanceCount",
    "max_instance_count",
    help="Maximal instance count",
    cls=common.GradientOption,
)
@click.option(
    "--scaleCooldownPeriod",
    "scale_cooldown_period",
    help="Scale cooldown period",
    cls=common.GradientOption,
)
@click.option(
    "--metric",
    "metrics",
    multiple=True,
    callback=validate_autoscaling_metric,
    help="Autoscaling metrics. Example: my_metric/targetAverage:21.37",
    cls=common.GradientOption,
)
@click.option(
    "--resource",
    "resources",
    multiple=True,
    callback=validate_autoscaling_resource,
    help="Autoscaling resources. Example: cpu/target:60",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def create_deployment(api_key, options_file, **kwargs):
    cliutils.validate_auth_options(kwargs)
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    del_if_value_is_none(kwargs)
    command = deployments_commands.CreateDeploymentCommand(api_key=api_key,
                                                           workspace_handler=get_workspace_handler(api_key))
    command.execute(**kwargs)


DEPLOYMENT_STATES_MAP = collections.OrderedDict(
    (
        ("BUILDING", "Building"),
        ("PROVISIONING", "Provisioning"),
        ("STARTING", "Starting"),
        ("RUNNING", "Running"),
        ("STOPPING", "Stopping"),
        ("STOPPED", "Stopped"),
        ("ERROR", "Error"),
    )
)


@deployments_group.command("list", help="List deployments with optional filtering")
@click.option(
    "--state",
    "state",
    type=ChoiceType(DEPLOYMENT_STATES_MAP, case_sensitive=False),
    help="Filter by deployment state",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="Use to filter by project ID",
    cls=common.GradientOption,
)
@click.option(
    "--modelId",
    "model_id",
    help="Use to filter by model ID",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    cls=common.GradientOption,
    help="Filter by tags. Multiple use"
)
@api_key_option
@common.options_file
def get_deployments_list(api_key, options_file, **filters):
    del_if_value_is_none(filters)
    command = deployments_commands.ListDeploymentsCommand(api_key=api_key)
    try:
        command.execute(**filters)
    except exceptions.ApplicationError as e:
        clilogger.CliLogger().error(e)


@deployments_group.command("start", help="Start deployment")
@click.option(
    "--id",
    "id_",
    required=True,
    help="Deployment ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def start_deployment(id_, options_file, api_key=None):
    command = deployments_commands.StartDeploymentCommand(api_key=api_key)
    command.execute(deployment_id=id_)


@deployments_group.command("stop", help="Stop deployment")
@click.option(
    "--id",
    "id_",
    required=True,
    help="Deployment ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def stop_deployment(id_, options_file, api_key=None):
    command = deployments_commands.StopDeploymentCommand(api_key=api_key)
    command.execute(deployment_id=id_)


@deployments_group.command("delete", help="Delete deployment")
@click.option(
    "--id",
    "id_",
    required=True,
    help="Deployment ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def delete_deployment(id_, options_file, api_key):
    command = deployments_commands.DeleteDeploymentCommand(api_key=api_key)
    command.execute(deployment_id=id_)


@deployments_group.command("update", help="Modify existing deployment")
@click.option(
    "--id",
    "deployment_id",
    required=True,
    help="ID of existing deployment",
    cls=common.GradientOption,
)
@click.option(
    "--deploymentType",
    "deployment_type",
    type=ChoiceType(DEPLOYMENT_TYPES_MAP, case_sensitive=False),
    help="Model deployment type",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="Project ID",
    cls=common.GradientOption,
)
@click.option(
    "--modelId",
    "model_id",
    help="ID of a trained model",
    cls=common.GradientOption,
)
@click.option(
    "--name",
    "name",
    help="Human-friendly name for new model deployment",
    cls=common.GradientOption,
)
@click.option(
    "--machineType",
    "machine_type",
    help="Type of machine for new deployment",
    cls=common.GradientOption,
)
@click.option(
    "--imageUrl",
    "image_url",
    help="Docker image for model serving",
    cls=common.GradientOption,
)
@click.option(
    "--instanceCount",
    "instance_count",
    type=int,
    help="Number of machine instances",
    cls=common.GradientOption,
)
@click.option(
    "--command",
    "command",
    help="Deployment command",
    cls=common.GradientOption,
)
@click.option(
    "--containerModelPath",
    "container_model_path",
    help="Container model path",
    cls=common.GradientOption,
)
@click.option(
    "--imageUsername",
    "image_username",
    help="Username used to access docker image",
    cls=common.GradientOption,
)
@click.option(
    "--imagePassword",
    "image_password",
    help="Password used to access docker image",
    cls=common.GradientOption,
)
@click.option(
    "--imageServer",
    "image_server",
    help="Docker image server",
    cls=common.GradientOption,
)
@click.option(
    "--containerUrlPath",
    "container_url_path",
    help="Container URL path",
    cls=common.GradientOption,
)
@click.option(
    "--method",
    "method",
    help="Method",
    cls=common.GradientOption,
)
@click.option(
    "--dockerArgs",
    "docker_args",
    type=json_string,
    help="JSON-style list of docker args",
    cls=common.GradientOption,
)
@click.option(
    "--env",
    "env",
    type=json_string,
    help="JSON-style environmental variables map",
    cls=common.GradientOption,
)
@click.option(
    "--apiType",
    "api_type",
    help="Type of API",
    cls=common.GradientOption,
)
@click.option(
    "--ports",
    "ports",
    help="Ports",
    cls=common.GradientOption,
)
@click.option(
    "--authUsername",
    "auth_username",
    help="Username",
    cls=common.GradientOption,
)
@click.option(
    "--authPassword",
    "auth_password",
    help="Password",
    cls=common.GradientOption,
)
@click.option(
    "--clusterId",
    "cluster_id",
    help="Cluster ID",
    cls=common.GradientOption,
)
@click.option(
    "--workspace",
    "workspace",
    help="Path to workspace directory, archive, S3 or git repository",
    cls=common.GradientOption,
)
@click.option(
    "--workspaceRef",
    "workspace_ref",
    help="Git commit hash, branch name or tag",
    cls=common.GradientOption,
)
@click.option(
    "--workspaceUsername",
    "workspace_username",
    metavar="<username>",
    help="Workspace username",
    cls=common.GradientOption,
)
@click.option(
    "--workspacePassword",
    "workspace_password",
    help="Workspace password",
    cls=common.GradientOption,
)
@click.option(
    "--minInstanceCount",
    "min_instance_count",
    help="Minimal instance count",
    cls=common.GradientOption,
)
@click.option(
    "--maxInstanceCount",
    "max_instance_count",
    help="Maximal instance count",
    cls=common.GradientOption,
)
@click.option(
    "--scaleCooldownPeriod",
    "scale_cooldown_period",
    help="Scale cooldown period",
    cls=common.GradientOption,
)
@click.option(
    "--metric",
    "metrics",
    multiple=True,
    callback=validate_autoscaling_metric,
    help="Autoscaling metrics. Example: my_metric/targetAverage:21.37",
    cls=common.GradientOption,
)
@click.option(
    "--resource",
    "resources",
    multiple=True,
    callback=validate_autoscaling_resource,
    help="Autoscaling resources. Example: cpu/target:60",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def update_deployment(deployment_id, api_key, options_file, **kwargs):
    del_if_value_is_none(kwargs)
    command = deployments_commands.UpdateDeploymentCommand(api_key=api_key,
                                                           workspace_handler=get_workspace_handler(api_key))
    command.execute(deployment_id, **kwargs)


@deployments_group.command("details", help="Get details of model deployment")
@click.option(
    "--id",
    "deployment_id",
    required=True,
    help="Deployment ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_deployment(deployment_id, api_key, options_file):
    command = deployments_commands.GetDeploymentDetails(api_key=api_key)
    command.execute(deployment_id)


@deployments_tags.command("add", help="Add tags to deployment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the deployment",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to deployment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to deployment",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def deployment_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = DeploymentAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@deployments_tags.command("remove", help="Remove tags from deployment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the deployment",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from deployment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from deployment",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def deployment_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = DeploymentRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@deployments_metrics.command(
    "get",
    short_help="Get model deployment metrics",
    help="Get model deployment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "deployment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the model deployment",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=str,
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help=("One or more metrics that you want to read: {}. Defaults to cpuPercentage and memoryUsage. To view available custom metrics, use command: `gradient deployments metrics list`".format(', '.join(map(str, constants.METRICS_MAP)))),
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_deployment_metrics(deployment_id, metrics_list, interval, start, end, options_file, api_key):
    command = GetDeploymentMetricsCommand(api_key=api_key)
    command.execute(deployment_id, start, end, interval, built_in_metrics=metrics_list)

@deployments_metrics.command(
    "list",
    short_help="List model deployment metrics",
    help="List model deployment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "deployment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the model deployment",
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_deployment_metrics(deployment_id, interval, start, end, options_file, api_key):
    command = ListDeploymentMetricsCommand(api_key=api_key)
    command.execute(deployment_id, start, end, interval)


@deployments_metrics.command(
    "stream",
    short_help="Watch live model deployment metrics",
    help="Watch live model deployment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "deployment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the model deployment",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=ChoiceType(constants.METRICS_MAP, case_sensitive=False),
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help="One or more metrics that you want to read. Defaults to cpuPercentage and memoryUsage",
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def stream_model_deployment_metrics(deployment_id, metrics_list, interval, options_file, api_key):
    command = StreamDeploymentMetricsCommand(api_key=api_key)
    command.execute(deployment_id=deployment_id, interval=interval, built_in_metrics=metrics_list)


@deployments_group.command("logs", help="List deployment logs")
@click.option(
    "--id",
    "deployment_id",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--line",
    "line",
    default=0,
    cls=common.GradientOption,
)
@click.option(
    "--limit",
    "limit",
    default=10000,
    cls=common.GradientOption,
)
@click.option(
    "--follow",
    "follow",
    default=False,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_logs(deployment_id, line, limit, follow, options_file, api_key=None):
    command = DeploymentLogsCommand(api_key=api_key)
    command.execute(deployment_id, line, limit, follow)



import functools

import click

from gradient import clilogger
from gradient.api_sdk import constants, workspace
from gradient.cli import common, validators
from gradient.cli.cli import cli
from gradient.cli.cli_types import json_string, ChoiceType
from gradient.cli.common import api_key_option, ClickGroup, validate_comma_split_option
from gradient.cli.utils.flag_with_value import GradientRegisterReaderOption, GradientRegisterWriterOption, \
    GradientRegisterWriterCommand
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.commands import experiments as experiments_commands
from gradient.commands.experiments import ExperimentAddTagsCommand, ExperimentRemoveTagsCommand, \
    GetExperimentMetricsCommand, ListExperimentMetricsCommand, StreamExperimentMetricsCommand

MULTI_NODE_CREATE_EXPERIMENT_COMMANDS = {
    constants.ExperimentType.GRPC_MULTI_NODE: experiments_commands.CreateMultiNodeExperimentCommand,
    constants.ExperimentType.MPI_MULTI_NODE: experiments_commands.CreateMpiMultiNodeExperimentCommand,
}

MULTI_NODE_RUN_EXPERIMENT_COMMANDS = {
    constants.ExperimentType.GRPC_MULTI_NODE: experiments_commands.CreateAndStartMultiNodeExperimentCommand,
    constants.ExperimentType.MPI_MULTI_NODE: experiments_commands.CreateAndStartMpiMultiNodeExperimentCommand,
}


def get_workspace_handler(api_key):
    logger_ = clilogger.CliLogger()
    workspace_handler = workspace.S3WorkspaceHandlerWithProgressbar(api_key=api_key, logger_=logger_,
                                                                    client_name=CLI_PS_CLIENT_NAME)
    return workspace_handler


@cli.group("experiments", help="Manage experiments", cls=ClickGroup)
def experiments_group():
    pass


@experiments_group.group("create", help="Create new experiment", cls=ClickGroup)
def create_experiment():
    pass


@experiments_group.group(name="run", help="Create and start new experiment", cls=ClickGroup)
def create_and_start_experiment():
    pass


@experiments_group.group(name="tags", help="Manage tags for experiment", cls=ClickGroup)
def experiments_tags():
    pass


@experiments_group.group(name="metrics", help="Read experiment metrics", cls=ClickGroup)
def experiments_metrics():
    pass


def common_experiments_create_options(f):
    options = [
        click.option(
            "--name",
            metavar="<name>",
            help="Name of new experiment",
            cls=common.GradientOption,
        ),
        click.option(
            "--ports",
            help="Port to use in new experiment",
            cls=common.GradientOption,
            ),
            click.option(
                "--workspace",
                "workspace",
                help="Path to workspace directory, archive, S3 or git repository",
                cls=common.GradientOption,
            ),
            click.option(
                "--workspaceRef",
                "workspace_ref",
                help="Git commit hash, branch name or tag",
                cls=common.GradientOption,
            ),
            click.option(
                "--workspaceUsername",
                "workspace_username",
                metavar="<username>",
                help="Workspace username",
                cls=common.GradientOption,
            ),
            click.option(
                "--workspacePassword",
                "workspace_password",
                help="Workspace password",
                cls=common.GradientOption,
            ),
        click.option(
            "--ignoreFiles",
            "ignore_files",
            help="Ignore certain files from uploading",
            cls=common.GradientOption,
        ),
        click.option(
            "--workingDirectory",
            "working_directory",
            help="Working directory for the experiment",
            cls=common.GradientOption,
        ),
        click.option(
            "--artifactDirectory",
            "artifact_directory",
            help="Artifacts directory",
            cls=common.GradientOption,
        ),
        click.option(
            "--clusterId",
            "cluster_id",
            metavar="<cluster ID>",
            help="Cluster ID",
            cls=common.GradientOption,
        ),
        click.option(
            "--experimentEnv",
            "experiment_env",
            type=json_string,
            help="Environment variables in a JSON",
            cls=common.GradientOption,
        ),
        click.option(
            "--projectId",
            "project_id",
            metavar="<project ID>",
            required=True,
            help="Project ID",
            cls=common.GradientOption,
        ),
        click.option(
            "--modelType",
            "model_type",
            metavar="<model type>",
            help="Model type",
            cls=common.GradientOption,
        ),
        click.option(
            "--modelPath",
            "model_path",
            metavar="<path>",
            help="Model path",
            cls=common.GradientOption,
        ),
        click.option(
            "--isPreemptible",
            "is_preemptible",
            type=bool,
            is_flag=True,
            help="Flag: is preemptible",
            cls=common.GradientOption,
        ),
        click.option(
            "--tag",
            "tags",
            multiple=True,
            help="One or many tags that you want to add to experiment",
            cls=common.GradientOption
        ),
        click.option(
            "--tags",
            "tags_comma",
            help="Separated by comma tags that you want add to experiment",
            cls=common.GradientOption
        )
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


def dataset_options(f):
    options = [
        click.option(
            "--datasetId",
            "dataset_id_list",
            metavar="<dateset id>",
            multiple=True,
            help="Dataset ID",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetUri",
            "dataset_uri_list",
            metavar="<dateset uri>",
            multiple=True,
            help="Url to S3 bucket with dataset",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetName",
            "dataset_name_list",
            multiple=True,
            metavar="<dateset name>",
            help="Name of dataset",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetAwsAccessKeyId",
            "dataset_access_key_id_list",
            multiple=True,
            metavar="<AWS access key>",
            help="S3 bucket's Access Key ID",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetAwsSecretAccessKey",
            "dataset_secret_access_key_list",
            multiple=True,
            help="S3 bucket's Secret Access Key",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetAwsEndpoint",
            "dataset_endpoint_list",
            multiple=True,
            help="S3 endpoint URL",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetVersionId",
            "dataset_version_id_list",
            metavar="<version ID>",
            multiple=True,
            help="S3 dataset's version ID",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetEtag",
            "dataset_etag_list",
            metavar="<etag>",
            multiple=True,
            help="S3 dataset's ETag",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetVolumeKind",
            "dataset_volume_kind_list",
            multiple=True,
            type=ChoiceType(constants.DATASET_VOLUME_KINDS, case_sensitive=False),
            help="S3 dataset's volume kind. If used, --datasetVolumeSize has to be set as well",
            cls=common.GradientDatasetOption,
        ),
        click.option(
            "--datasetVolumeSize",
            "dataset_volume_size_list",
            multiple=True,
            help="S3 dataset's volume size",
            cls=common.GradientDatasetOption,
        ),
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


def common_experiment_create_multi_node_options(f):
    options = [
        click.option(
            "--experimentType",
            "experiment_type_id",
            type=ChoiceType(constants.MULTI_NODE_EXPERIMENT_TYPES_MAP, case_sensitive=False),
            required=True,
            help="Experiment Type",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerContainer",
            "worker_container",
            metavar="<container>",
            required=True,
            help="Worker container",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerMachineType",
            "worker_machine_type",
            metavar="<machine type>",
            required=True,
            help="Worker machine type",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerCommand",
            "worker_command",
            metavar="<command>",
            required=True,
            help="Worker command",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerCount",
            "worker_count",
            type=int,
            required=True,
            help="Worker count",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerContainer",
            "parameter_server_container",
            metavar="<container>",
            help="Parameter server container (GRPC only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerMachineType",
            "parameter_server_machine_type",
            metavar="<machine type>",
            help="Parameter server machine type (GRPC only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerCommand",
            "parameter_server_command",
            metavar="<command>",
            help="Parameter server command (GRPC only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerCount",
            "parameter_server_count",
            type=int,
            help="Parameter server count (GRPC only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterContainer",
            "master_container",
            metavar="<container>",
            help="Master container (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterMachineType",
            "master_machine_type",
            metavar="<machine type>",
            help="Master machine type (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterCount",
            "master_count",
            help="Master count (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterCommand",
            "master_command",
            metavar="<command>",
            help="Master command (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerContainerUser",
            "worker_container_user",
            help="Worker container user",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerRegistryUsername",
            "worker_registry_username",
            help="Worker container registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerRegistryPassword",
            "worker_registry_password",
            metavar="<password>",
            help="Worker registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerRegistryUrl",
            "worker_registry_url",
            metavar="<registry url>",
            help="Worker registry URL",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerContainerUser",
            "parameter_server_container_user",
            help="Parameter server container user",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerRegistryUsername",
            "parameter_server_registry_username",
            help="Parameter server registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerRegistryPassword",
            "parameter_server_registry_password",
            metavar="<password>",
            help="Parameter server registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--parameterServerRegistryUrl",
            "parameter_server_registry_url",
            metavar="<registry url>",
            help="Parameter server registry URL",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterContainerUser",
            "master_container_user",
            help="Master container user (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterRegistryUsername",
            "master_registry_username",
            metavar="<username>",
            help="Master registry username (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterRegistryPassword",
            "master_registry_password",
            metavar="<password>",
            help="Master registry password (MPI only)",
            cls=common.GradientOption,
        ),
        click.option(
            "--masterRegistryUrl",
            "master_registry_url",
            metavar="<registry url>",
            help="Master registry URL (MPI only)",
            cls=common.GradientOption
        ),
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


def common_experiments_create_single_node_options(f):
    options = [
        click.option(
            "--container",
            required=True,
            help="Container",
            cls=common.GradientOption,
        ),
        click.option(
            "--machineType",
            "machine_type",
            required=True,
            metavar="<machine type>",
            help="Machine type",
            cls=common.GradientOption,
        ),
        click.option(
            "--command",
            required=True,
            metavar="<command>",
            help="Container entrypoint command",
            cls=common.GradientOption,
        ),
        click.option(
            "--containerUser",
            "container_user",
            help="Container user",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryUsername",
            "registry_username",
            help="Registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryPassword",
            "registry_password",
            metavar="<password>",
            help="Registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryUrl",
            "registry_url",
            metavar="<registry url>",
            help="Registry URL",
            cls=common.GradientOption,
        ),
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


def tensorboard_option(f):
    options = [
        click.option(
            "--tensorboard",
            is_flag=True,
            # default=experiments_commands.NoTensorboardId,
            help="Creates new tensorboard for this experiment",
            cls=GradientRegisterReaderOption,
        ),
        click.option(
            "--tensorboard_set",
            help="Add to existing tensorboard",
            cls=GradientRegisterWriterOption,
            metavar='<tensorboard ID>'
        ),
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


def parse_tensorboard_options(tensorboard, tensorboard_set):
    """
    :param str|bool tensorboard:
    :param str|None tensorboard_set:
    :rtype: str|bool
    """
    if tensorboard is True:
        return True

    if tensorboard_set:
        return tensorboard_set
    else:
        return False


@create_experiment.command(name="multinode", help="Create multi node experiment", cls=GradientRegisterWriterCommand)
@common_experiments_create_options
@common_experiment_create_multi_node_options
@dataset_options
@tensorboard_option
@api_key_option
@common.options_file
def create_multi_node(api_key, tensorboard, tensorboard_set, options_file, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    add_to_tensorboard = parse_tensorboard_options(tensorboard, tensorboard_set)

    validators.validate_multi_node(kwargs)
    common.del_if_value_is_none(kwargs)
    experiment_type = kwargs.get('experiment_type_id')
    command_class = MULTI_NODE_CREATE_EXPERIMENT_COMMANDS.get(experiment_type)
    command = command_class(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    command.execute(kwargs, add_to_tensorboard=add_to_tensorboard)


@create_experiment.command(name="singlenode", help="Create single node experiment", cls=GradientRegisterWriterCommand)
@common_experiments_create_options
@common_experiments_create_single_node_options
@dataset_options
@tensorboard_option
@api_key_option
@common.options_file
def create_single_node(api_key, tensorboard, tensorboard_set, options_file, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    add_to_tensorboard = parse_tensorboard_options(tensorboard, tensorboard_set)

    common.del_if_value_is_none(kwargs)

    command = experiments_commands.CreateSingleNodeExperimentCommand(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    command.execute(kwargs, add_to_tensorboard=add_to_tensorboard)


@create_and_start_experiment.command(name="multinode", help="Create and start new multi node experiment",
                                     cls=GradientRegisterWriterCommand)
@common_experiments_create_options
@common_experiment_create_multi_node_options
@click.option(
    "--no-logs",
    "show_logs",
    is_flag=True,
    flag_value=False,
    default=True,
    help="Don't show logs. Only create, start and exit",
)
@dataset_options
@tensorboard_option
@api_key_option
@common.options_file
@click.pass_context
def create_and_start_multi_node(ctx, api_key, show_logs, tensorboard, tensorboard_set, options_file, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    add_to_tensorboard = parse_tensorboard_options(tensorboard, tensorboard_set)

    validators.validate_multi_node(kwargs)
    common.del_if_value_is_none(kwargs)

    experiment_type = kwargs.get('experiment_type_id')
    command_class = MULTI_NODE_RUN_EXPERIMENT_COMMANDS.get(experiment_type)

    command = command_class(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    experiment_id = command.execute(kwargs, add_to_tensorboard=add_to_tensorboard)
    if experiment_id and show_logs:
        ctx.invoke(list_logs, experiment_id=experiment_id, line=1, limit=100, follow=True, api_key=api_key)


@create_and_start_experiment.command(name="singlenode", help="Create and start new single node experiment",
                                     cls=GradientRegisterWriterCommand)
@common_experiments_create_options
@common_experiments_create_single_node_options
@click.option(
    "--no-logs",
    "show_logs",
    is_flag=True,
    flag_value=False,
    default=True,
    help="Don't show logs. Only create, start and exit",
)
@dataset_options
@tensorboard_option
@api_key_option
@common.options_file
@click.pass_context
def create_and_start_single_node(ctx, api_key, show_logs, tensorboard, tensorboard_set, options_file,
                                 **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    add_to_tensorboard = parse_tensorboard_options(tensorboard, tensorboard_set)

    common.del_if_value_is_none(kwargs)

    command = experiments_commands.CreateAndStartSingleNodeExperimentCommand(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    experiment_id = command.execute(kwargs, add_to_tensorboard=add_to_tensorboard)
    if experiment_id and show_logs:
        ctx.invoke(list_logs, experiment_id=experiment_id, line=1, limit=100, follow=True, api_key=api_key)


@experiments_group.command("start", help="Start experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--logs",
    "show_logs",
    is_flag=True,
    help="Show logs",
)
@api_key_option
@common.options_file
@click.pass_context
def start_experiment(ctx, id, show_logs, api_key, options_file):
    command = experiments_commands.StartExperimentCommand(api_key=api_key)
    command.execute(id)

    if show_logs:
        ctx.invoke(list_logs, experiment_id=id, line=1, limit=100, follow=True, api_key=api_key)


@experiments_group.command("stop", help="Stop experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@api_key_option
@common.options_file
def stop_experiment(id, api_key, options_file):
    command = experiments_commands.StopExperimentCommand(api_key=api_key)
    command.execute(id)


@experiments_group.command("list", help="List experiments")
@click.option(
    "--projectId",
    "-p",
    "project_ids",
    multiple=True,
    metavar='<project ID>',
    help="Filter by project IDs. Multiple use",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    cls=common.GradientOption,
    help="Filter by tags. Multiple use"
)
@click.option(
    "--limit",
    "-l",
    "exp_limit",
    default=20,
    help="Limit listed experiments per page",
    cls=common.GradientOption,
)
@click.option(
    "--offset",
    "-o",
    "exp_offset",
    default=0,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_experiments(project_ids, api_key, exp_limit, exp_offset, tags, options_file):
    command = experiments_commands.ListExperimentsCommand(api_key=api_key)

    res = command.execute(project_id=project_ids, limit=exp_limit, offset=exp_offset, tags=tags)
    for experiments_str, next_iteration in res:
        click.echo(experiments_str)
        if next_iteration:
            click.confirm("Do you want to continue?", abort=True)


@experiments_group.command("details", help="Show detail of an experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@api_key_option
@common.options_file
def get_experiment_details(id, options_file, api_key):
    command = experiments_commands.GetExperimentCommand(api_key=api_key)
    command.execute(id)


@experiments_group.command("logs", help="List experiment logs")
@click.option(
    "--id",
    "experiment_id",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--line",
    "line",
    required=False,
    default=0,
    cls=common.GradientOption,
)
@click.option(
    "--limit",
    "limit",
    required=False,
    default=10000,
    cls=common.GradientOption,
)
@click.option(
    "--follow",
    "follow",
    required=False,
    default=False,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_logs(experiment_id, line, limit, follow, options_file, api_key=None):
    command = experiments_commands.ExperimentLogsCommand(api_key=api_key)
    command.execute(experiment_id, line, limit, follow)


@experiments_group.command("delete", help="Delete an experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@api_key_option
@common.options_file
def delete_experiment(id, options_file, api_key):
    command = experiments_commands.DeleteExperimentCommand(api_key=api_key)
    command.execute(id)


@experiments_tags.command("add", help="Add tags to experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to experiment",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def experiment_add_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = ExperimentAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@experiments_tags.command("remove", help="Remove tags from experiment")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from experiment",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def experiment_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = ExperimentRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@experiments_metrics.command(
    "get",
    short_help="Get experiment metrics",
    help="Get experiment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "experiment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=str,
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help=("One or more metrics that you want to read: {}. Defaults to cpuPercentage and memoryUsage. To view available custom metrics, use command: `gradient experiments metrics list`".format(', '.join(map(str, constants.METRICS_MAP)))),
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_experiment_metrics(experiment_id, metrics_list, interval, start, end, options_file, api_key):
    command = GetExperimentMetricsCommand(api_key=api_key)
    command.execute(experiment_id, start, end, interval, built_in_metrics=metrics_list)


@experiments_metrics.command(
    "list",
    short_help="List experiment metrics",
    help="List experiment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "experiment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_experiment_metrics(experiment_id, interval, start, end, options_file, api_key):
    command = ListExperimentMetricsCommand(api_key=api_key)
    command.execute(experiment_id, start, end, interval)


@experiments_metrics.command(
    "stream",
    short_help="Watch live experiment metrics",
    help="Watch live experiment metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "experiment_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the experiment",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=ChoiceType(constants.METRICS_MAP, case_sensitive=False),
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help="One or more metrics that you want to read. Defaults to cpuPercentage and memoryUsage",
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_experiment_metrics_stream(experiment_id, metrics_list, interval, options_file, api_key):
    command = StreamExperimentMetricsCommand(api_key=api_key)
    command.execute(experiment_id=experiment_id, interval=interval, built_in_metrics=metrics_list)



import mock
from click.testing import CliRunner

from gradient.api_sdk.clients import http_client
from gradient.cli import cli
from tests import example_responses, MockResponse

EXPECTED_HEADERS = http_client.default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"

URL = "https://api.paperspace.io"


class TestListStorageProviders(object):
    COMMAND = ["storageProviders", "list"]

    JSON = example_responses.LIST_STORAGE_PROVIDERS_RESPONSE

    STDOUT = """+-------+-----------------+------+---------------------------------+
| Name  | ID              | Type | Config                          |
+-------+-----------------+------+---------------------------------+
| test1 | spltautet072md4 | s3   | accessKey: AKIBAEG7J3OJ24XAV33B |
|       |                 |      | bucket: bucket                  |
|       |                 |      | secretAccessKey: ********       |
+-------+-----------------+------+---------------------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_list_storage_providers(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND)

        assert self.STDOUT == result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/storageProviders",
            headers=EXPECTED_HEADERS,
            json=None,
            params={
                "filter[limit]": 21,
                "filter[skip]": 0,
                "filter[order][]": "name ASC",
            }
        )


class TestShowStorageProviderDetails(object):
    COMMAND = ["storageProviders", "details"]

    JSON = example_responses.SHOW_STORAGE_PROVIDER_DETAILS_RESPONSE

    STDOUT = """+--------+---------------------------------+
| Name   | test1                           |
+--------+---------------------------------+
| ID     | spltautet072md4                 |
| Type   | s3                              |
| Config | accessKey: AKIBAEG7J3OJ24XAV33B |
|        | bucket: bucket                  |
|        | secretAccessKey: ********       |
+--------+---------------------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_show_storage_provider_details(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=spltautet072md4"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/storageProviders/spltautet072md4",
            headers=EXPECTED_HEADERS,
            json=None,
            params=None
        )


class TestCreateStorageProvider(object):
    COMMAND = ["storageProviders", "create"]

    JSON = example_responses.CREATE_STORAGE_PROVIDER_RESPONSE

    STDOUT = """Created new storage provider with id: spltautet072md4
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_create_storage_provider(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + [
            "s3", "--name=test1", "--bucket=bucket", "--accessKey=AKIBAEG7J3OJ24XAV33B", "--secretAccessKey=bla"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/storageProviders",
            json={
                "name": "test1",
                "type": "s3",
                "config": {"bucket": "bucket", "accessKey": "AKIBAEG7J3OJ24XAV33B", "secretAccessKey": "bla"},
            },
            params=None,
            headers=EXPECTED_HEADERS,
            files=None,
            data=None
        )


class TestUpdateStorageProvider(object):
    COMMAND = ["storageProviders", "update"]

    JSON = example_responses.UPDATE_STORAGE_PROVIDER_RESPONSE

    STDOUT = """Updated storage provider
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_update_storage_provider(self, method):
        method.return_value = MockResponse(self.JSON)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["s3", "--id=spltautet072md4", "--name=test2"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/storageProviders/spltautet072md4",
            json={"name": "test2"},
            params=None,
            headers=EXPECTED_HEADERS,
            files=None,
            data=None
        )


class TestDeleteStorageProvider(object):
    COMMAND = ["storageProviders", "delete"]

    STDOUT = """Deleted storage provider: spltautet072md4
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_delete_storage_provider(self, method):
        method.return_value = MockResponse(status_code=204)

        result = CliRunner().invoke(cli.cli, self.COMMAND + ["--id=spltautet072md4"])

        assert self.STDOUT in result.output, result.exc_info
        method.assert_called_once_with(
            URL + "/storageProviders/spltautet072md4",
            params=None,
            headers=EXPECTED_HEADERS,
            json=None
        )



from collections import OrderedDict

import mock
import pytest

import gradient.api_sdk.utils
from gradient import cliutils, exceptions
from gradient.api_sdk.sdk_exceptions import WrongPathError
from gradient.api_sdk.utils import PathParser
from gradient.cli.common import validate_comma_split_option

output_response = ""


class TestPrintDictRecursive(object):
    def test_json_print(self):
        global output_response
        output_response = ""

        def log_to_var(message):
            global output_response
            output_response = "{}{}\n".format(output_response, message)

        logger_ = mock.MagicMock()
        logger_.log = log_to_var

        input_dict = {
            "foo": {
                'bar': {
                    "baz": "faz"
                }
            }
        }
        expected_string = """foo:
  bar:
    baz:
      faz
"""

        gradient.api_sdk.utils.print_dict_recursive(OrderedDict(input_dict), logger_)

        assert output_response == expected_string


class TestPathParser(object):
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("os.path.isdir", return_value=True)
    def test_should_return_local_dir_type_when_given_directory_exists(self, _, __):
        path_str = "/home/usr/some/path"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.LOCAL_DIR

    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=True)
    def test_should_return_local_archive_type_when_archive_exists(self, _, __, ___):
        path_str = "/home/usr/some/path.zip"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.LOCAL_FILE

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_git_url_type_when_git_ssh_url_was_given(self, _, __, ___):
        path_str = "git@github.com:Paperspace/gradient-cli.git"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.GIT_URL

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_git_url_type_when_git_http_url_was_given(self, _, __, ___):
        path_str = "https://github.com/Paperspace/gradient-cli.git"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.GIT_URL

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_git_url_type_when_git_http_url_was_given_with_git_prefix_and_no_extension(self, _, __, ___):
        path_str = "GIT:https://github.com/Paperspace/gradient-cli"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.GIT_URL

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_s3_type_when_s3_url_was_given(self, _, __, ___):
        path_str = "S3:some/path"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.S3_URL

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_raise_exception_when_local_path_was_given_but_does_not_exist(self, _, __, ___):
        path_str = "/home/usr/some/path.zip"

        with pytest.raises(WrongPathError):
            PathParser.parse_path(path_str)

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_http_for_http_url(self, _, __, ___):
        path_str = "http://pastebin.com/raw/2VPPwrnH"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.HTTP_URL

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("os.path.isfile", return_value=False)
    def test_should_return_http_for_https_url(self, _, __, ___):
        path_str = "https://pastebin.com/raw/2VPPwrnH"

        path_type = PathParser.parse_path(path_str)

        assert path_type == PathParser.HTTP_URL


class TestValidateAuthOptions(object):
    def test_should_not_raise_exception_when_only_generate_auth_was_set(self):
        kwargs = {"auth_username": None, "auth_password": None, "generate_auth": True}
        cliutils.validate_auth_options(kwargs)

    def test_should_not_raise_exception_when_username_and_password_was_set(self):
        kwargs = {"auth_username": "username", "auth_password": "password", "generate_auth": False}
        cliutils.validate_auth_options(kwargs)

    def test_should_not_raise_exception_when_username_was_set_but_no_password(self):
        kwargs = {"auth_username": "username", "auth_password": None, "generate_auth": False}
        pytest.raises(exceptions.ApplicationError, cliutils.validate_auth_options, kwargs)

    def test_should_not_raise_exception_when_all_values_were_set(self):
        kwargs = {"auth_username": "username", "auth_password": "password", "generate_auth": True}
        pytest.raises(exceptions.ApplicationError, cliutils.validate_auth_options, kwargs)


class TestCommonFunction(object):
    @pytest.mark.parametrize("comma_value, value, expected_result", [
        pytest.param(None, None, None, id="None passed in both argument"),
        pytest.param(None, ['test'], ['test'], id="None passed as comma separated value"),
        pytest.param("test", None, ['test'], id="None passed as list value"),
        pytest.param("test0", ["test1"], ["test0", "test1"], id="Pass single values for arguments"),
        pytest.param(
            "test0,test1", ["test2"], ["test0", "test1", "test2"],
            id="Pass more values separated by comma and single element in list"
        ),
        pytest.param(
            "test0 ,test1, test2 , test3", ["test4"], ["test0", "test1", "test2", "test3", "test4"],
            id="Pass more values separated by comma in different style and single element in list"
        ),
        pytest.param(
            "test0\t,test1,\ntest2\t,\ttest3\n", ["test4"], ["test0", "test1", "test2", "test3", "test4"],
            id="Pass more values separated by comma with more type of white chars and single element in list"
        ),
    ])
    def test_validate_comma_split_option(self, comma_value, value, expected_result):
        if expected_result:
            expected_result.sort()
        result = validate_comma_split_option(comma_value, value)
        if result:
            result.sort()
        assert result == expected_result



"""
Each command should be tested with following scenarios of execution:
  - with required parameters only
  - with all parameters used (if possible as some may be mutually exclusive)
  - with --apiKey parameter used
  - using --optionsFile if available
  - using wrong apiKey - with patched response from API
  - API returned error with message in JSON
  - API returned error without JSON
  - check if mutual exclusion of parameters works as expected
"""



import abc

import halo
import six

from gradient import api_sdk, exceptions, cli_constants
from gradient.api_sdk import sdk_exceptions
from gradient.api_sdk.s3_downloader import ModelFilesDownloader
from gradient.commands.common import BaseCommand, ListCommandMixin, DetailsCommandMixin
from gradient.exceptions import ApplicationError


@six.add_metaclass(abc.ABCMeta)
class GetModelsClientMixin:
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.ModelsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=cli_constants.CLI_PS_CLIENT_NAME,
        )
        return client


class ListModelsCommand(GetModelsClientMixin, ListCommandMixin, BaseCommand):
    def _get_instances(self, kwargs):
        try:
            instances = self.client.list(**kwargs)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances

    def _get_table_data(self, models):
        data = [("Name", "ID", "Model Type", "Project ID", "Experiment ID")]
        for model in models:
            name = model.name
            id_ = model.id
            project_id = model.project_id
            experiment_id = model.experiment_id
            model_type = model.model_type
            data.append((name, id_, model_type, project_id, experiment_id))

        return data


class DeleteModelCommand(GetModelsClientMixin, BaseCommand):
    def execute(self, model_id, *args, **kwargs):
        self.client.delete(model_id)
        self.logger.log("Model deleted")


class CreateModel(GetModelsClientMixin, BaseCommand):
    SPINNER_MESSAGE = "Creating model"

    def execute(self, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            model_id = self.client.create(**kwargs)

        self.logger.log("Model created with ID: {}".format(model_id))


class UploadModel(GetModelsClientMixin, BaseCommand):
    SPINNER_MESSAGE = "Uploading model"

    def execute(self, **kwargs):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            model_id = self.client.upload(**kwargs)

        self.logger.log("Model uploaded with ID: {}".format(model_id))


class GetModelCommand(DetailsCommandMixin, GetModelsClientMixin, BaseCommand):
    def _get_table_data(self, instance):
        """
        :param api_sdk.Model instance:
        """
        tags_string = ", ".join(instance.tags)

        data = (
            ("ID", instance.id),
            ("Name", instance.name),
            ("Project ID", instance.project_id),
            ("Experiment ID", instance.experiment_id),
            ("Model Type", instance.model_type),
            ("URL", instance.url),
            ("Deployment State", instance.deployment_state),
            ("Tags", tags_string),
        )
        return data


class DownloadModelFiles(GetModelsClientMixin, BaseCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Downloading files..."

    def execute(self, model_id, destination_directory):
        model_files_downloader = ModelFilesDownloader(
            self.api_key,
            logger=self.logger,
            ps_client_name=cli_constants.CLI_PS_CLIENT_NAME,
        )
        try:
            model_files_downloader.download(model_id, destination_directory)
        except OSError as e:
            raise ApplicationError(e)


class MLModelAddTagsCommand(GetModelsClientMixin, BaseCommand):
    def execute(self, ml_model_id, *args, **kwargs):
        self.client.add_tags(ml_model_id, **kwargs)
        self.logger.log("Tags added to ml model")


class MLModelRemoveTagsCommand(GetModelsClientMixin, BaseCommand):
    def execute(self, ml_model_id, *args, **kwargs):
        self.client.remove_tags(ml_model_id, **kwargs)
        self.logger.log("Tags removed from ml model")



from .common import AlterResource, ListResources
from .. import config


class UpdateTagRepository(AlterResource):
    def get_request_url(self, **kwargs):
        return "/entityTags/updateTags"

    def _get_api_url(self, **kwargs):
        return config.config.CONFIG_HOST

    def _get_request_json(self, instance_dict):
        return {
            "entity": instance_dict.get("entity"),
            "entityId": instance_dict.get("entity_id"),
            "tags": instance_dict.get("tags")
        }

    def update(self, **kwargs):
        self._run(**kwargs)

    def _send_request(self, client, url, json_data=None):
        response = client.put(url, json=json_data)
        return response


class ListTagRepository(ListResources):
    def get_request_url(self, **kwargs):
        return "/entityTags/getTags"

    def _get_api_url(self, **kwargs):
        return config.config.CONFIG_HOST

    def _get_request_params(self, kwargs):
        return {
            "entity": kwargs.get("entity"),
            "entityIds": kwargs.get("entity_ids")
        }

    def _parse_objects(self, data, **kwargs):
        instances = []
        instance_dicts = self._get_instance_dicts(data, **kwargs)
        for entity in instance_dicts:
            tags = self._parse_object(instance_dicts.get(entity))
            instances.append({entity: tags})

        return instances

    def _parse_object(self, entity_tags):
        tags = []
        for tag in entity_tags:
            tags.append(tag.get("tag").get("name"))
        return tags



from .artifact import Artifact
from .cluster import Cluster
from .dataset import Dataset, DatasetRef
from .dataset_tag import DatasetTag, DatasetVersionSummary
from .dataset_version import DatasetVersion, DatasetVersionPreSignedS3Call, DatasetVersionPreSignedURL, \
    DatasetVersionTagSummary
from .deployment import Deployment, AutoscalingDefinition, AutoscalingMetric
from .experiment import BaseExperiment, MultiNodeExperiment, SingleNodeExperiment, MpiMultiNodeExperiment
from .experiment_dataset import ExperimentDataset, VolumeOptions
from .hyperparameter import Hyperparameter
from .job import Job, JobDataset
from .log import LogRow
from .machine import Machine, MachineEvent, MachineUtilization
from .model import Model, ModelFile
from .notebook import Notebook, NotebookStart
from .pagination import Pagination
from .project import Project
from .secret import Secret
from .storage_provider import StorageProvider
from .tag import Tag
from .tensorboard import Instance, Tensorboard
from .vm_type import VmType, VmTypeGpuModel
from .workflows import Workflow, WorkflowRun, WorkflowSpec



import attr


@attr.s
class VmTypeGpuModel(object):
    label = attr.ib(type=str, default=None)
    model = attr.ib(type=str, default=None)
    memory_in_bytes = attr.ib(type=int, default=None)


@attr.s
class VmType(object):
    label = attr.ib(type=str, default=None)
    kind = attr.ib(type=str, default=None)
    cpu_count = attr.ib(type=int, default=None)
    ram_in_bytes = attr.ib(type=int, default=None)
    gpu_count = attr.ib(type=int, default=None)
    gpu_model = attr.ib(type=VmTypeGpuModel, default=None)
    is_preemptible = attr.ib(type=bool, default=None)
    deployment_type = attr.ib(type=str, default=None)
    deployment_size = attr.ib(type=str, default=None)
    clusters = attr.ib(type=list, factory=list)



from .clients import *
from .constants import *
from .models import *
from .repositories import *
from .s3_downloader import JobArtifactsDownloader
from .s3_uploader import ExperimentFileUploader, ExperimentWorkspaceDirectoryUploader
from .archivers import ZipArchiver
from .sdk_exceptions import *



import marshmallow

from .base import BaseSchema

from .. import models
from ..sdk_exceptions import GradientSdkError
from ..serializers import SingleNodeExperimentSchema, MultiNodeExperimentSchema, HyperparameterSchema, \
    MpiMultiNodeExperimentSchema


EXPERIMENT_ID_TO_EXPERIMENT_SERIALIZER_MAPPING = {
    1: SingleNodeExperimentSchema,
    2: MultiNodeExperimentSchema,
    3: MpiMultiNodeExperimentSchema,
    4: HyperparameterSchema,
}


def get_serializer_for_experiment(experiment_dict):
    """
    :param dict experiment_dict:
    :rtype: serializers.BaseExperimentSchema
    """
    experiment_type_id = experiment_dict["experimentTypeId"]

    try:
        serializer = EXPERIMENT_ID_TO_EXPERIMENT_SERIALIZER_MAPPING[experiment_type_id]
    except KeyError as e:
        raise GradientSdkError("No experiment type with ID: {}".format(str(e)))

    return serializer


def paginate_schema(schema):
    class PaginationSchema(BaseSchema):
        MODEL = models.Pagination
        data = marshmallow.fields.Nested(schema, many=True, dump_only=True)
        start_after = marshmallow.fields.String(dump_to='startAfter', load_from='startAfter')

    return PaginationSchema()